"""Model exporting tool."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pathlib import Path

from otx.api.configuration.helper import create
from otx.api.entities.model import ModelEntity, ModelOptimizationType
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.cli.manager import ConfigManager
from otx.cli.utils.importing import get_impl_class
from otx.cli.utils.io import read_binary, read_label_schema, save_model_data
from otx.cli.utils.nncf import is_checkpoint_nncf
from otx.cli.utils.parser import get_parser_and_hprams_data


def get_args():
    """Parses command line arguments."""
    parser, _, _ = get_parser_and_hprams_data()

    parser.add_argument(
        "--load-weights",
        help="Load model weights from previously saved checkpoint.",
    )
    parser.add_argument(
        "--save-model-to",
        help="Location where exported model will be stored.",
    )
    parser.add_argument(
        "--dump_features",
        action="store_true",
        help="Whether to return feature vector and saliency map for explanation purposes.",
    )

    return parser.parse_args()


def main():
    """Main function that is used for model exporting."""
    args = get_args()
    config_manager = ConfigManager(args, mode="eval")
    # Auto-Configuration for model template
    config_manager.configure_template()

    # Load template.yaml file.
    template = config_manager.template

    # Get class for Task.
    if not args.load_weights and config_manager.check_workspace():
        args.load_weights = str(config_manager.workspace_root / "models/weights.pth")
    is_nncf = is_checkpoint_nncf(args.load_weights)
    task_class = get_impl_class(template.entrypoints.nncf if is_nncf else template.entrypoints.base)

    # Get hyper parameters schema.
    hyper_parameters = create(template.hyper_parameters.data)
    assert hyper_parameters

    environment = TaskEnvironment(
        model=None,
        hyper_parameters=hyper_parameters,
        label_schema=read_label_schema(args.load_weights),
        model_template=template,
    )

    model_adapters = {"weights.pth": ModelAdapter(read_binary(args.load_weights))}
    model = ModelEntity(
        configuration=environment.get_model_configuration(),
        model_adapters=model_adapters,
        train_dataset=None,
        optimization_type=ModelOptimizationType.NNCF if is_nncf else ModelOptimizationType.NONE,
    )
    environment.model = model

    task = task_class(task_environment=environment)

    exported_model = ModelEntity(None, environment.get_model_configuration())

    task.export(ExportType.OPENVINO, exported_model, args.dump_features)

    if "save_model_to" not in args or not args.save_model_to:
        args.save_model_to = str(config_manager.workspace_root / "model-exported")
    Path(args.save_model_to).mkdir(exist_ok=True, parents=True)
    save_model_data(exported_model, args.save_model_to)

    return dict(retcode=0, template=template.name)


if __name__ == "__main__":
    main()



"""Implementation of MobileNetV3.

Original papers:
- 'Searching for MobileNetV3,' https://arxiv.org/abs/1905.02244.
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name, too-many-arguments, unused-argument
# pylint: disable=too-many-locals, too-many-instance-attributes, abstract-method

import math
import os

import torch.nn.functional as F
from mmcls.models.utils import make_divisible
from mmcv.runner import load_checkpoint
from torch import nn

from otx.mpa.utils.logger import get_logger

from ..builder import BACKBONES

logger = get_logger()

pretrained_root = "https://github.com/d-li14/mobilenetv3.pytorch/blob/master/pretrained/"
pretrained_urls = {
    "mobilenetv3_small": pretrained_root + "mobilenetv3-small-55df8e1f.pth?raw=true",
    "mobilenetv3_large": pretrained_root + "mobilenetv3-large-1cd25616.pth?raw=true",
    "mobilenetv3_large_075": pretrained_root + "mobilenetv3-large-0.75-9632d2a8.pth?raw=true",
}


class ModelInterface(nn.Module):
    """Model Interface."""

    def __init__(
        self,
        classification=False,
        contrastive=False,
        pretrained=False,
        loss="softmax",
        **kwargs,
    ):
        super().__init__()

        self.classification = classification
        self.contrastive = contrastive
        self.pretrained = pretrained
        self.classification_classes = {}
        self.loss = loss
        self.is_ie_model = False
        if loss == "am_softmax":
            self.use_angle_simple_linear = True
        else:
            self.use_angle_simple_linear = False

    @staticmethod
    def _glob_feature_vector(x, mode, reduce_dims=True):
        if mode == "avg":
            out = F.adaptive_avg_pool2d(x, 1)
        elif mode == "max":
            out = F.adaptive_max_pool2d(x, 1)
        elif mode == "avg+max":
            avg_pool = F.adaptive_avg_pool2d(x, 1)
            max_pool = F.adaptive_max_pool2d(x, 1)
            out = avg_pool + max_pool
        else:
            raise ValueError(f"Unknown pooling mode: {mode}")

        if reduce_dims:
            return out.view(x.size(0), -1)
        return out


class HSigmoid(nn.Module):
    """Approximated sigmoid function, so-called hard-version of sigmoid from 'Searching for MobileNetV3,'.

    https://arxiv.org/abs/1905.02244.
    """

    def forward(self, x):
        """Forward."""
        return F.relu6(x + 3.0, inplace=True) / 6.0


class HSwish(nn.Module):
    """H-Swish activation function from 'Searching for MobileNetV3,'.

    https://arxiv.org/abs/1905.02244.

    Parameters:
        inplace : bool, Whether to use inplace version of the module.
    """

    def __init__(self, inplace=False):
        super().__init__()
        self.inplace = inplace

    def forward(self, x):
        """Forward."""
        return x * F.relu6(x + 3.0, inplace=self.inplace) / 6.0


class SELayer(nn.Module):
    """SE layer."""

    def __init__(self, channel, reduction=4):
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, make_divisible(channel // reduction, 8)),
            nn.ReLU(inplace=True),
            nn.Linear(make_divisible(channel // reduction, 8), channel),
            HSigmoid(),
        )

    def forward(self, x):
        """Forward."""

        # with no_nncf_se_layer_context():
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y


def conv_3x3_bn(inp, oup, stride, IN_conv1=False):
    """Conv 3x3 layer with batch-norm."""

    return nn.Sequential(
        nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
        nn.BatchNorm2d(oup) if not IN_conv1 else nn.InstanceNorm2d(oup, affine=True),
        HSwish(),
    )


def conv_1x1_bn(inp, oup, loss="softmax"):
    """Conv 1x1 layer with batch-norm."""

    return nn.Sequential(
        nn.Conv2d(inp, oup, 1, 1, 0, bias=False),
        nn.BatchNorm2d(oup),
        HSwish() if loss == "softmax" else nn.PReLU(),
    )


class InvertedResidual(nn.Module):
    """Inverted residual."""

    def __init__(self, inp, hidden_dim, oup, kernel_size, stride, use_se, use_hs):
        super().__init__()
        assert stride in [1, 2]

        self.identity = stride == 1 and inp == oup

        if inp == hidden_dim:
            self.conv = nn.Sequential(
                # dw
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    kernel_size,
                    stride,
                    (kernel_size - 1) // 2,
                    groups=hidden_dim,
                    bias=False,
                ),
                nn.BatchNorm2d(hidden_dim),
                HSwish() if use_hs else nn.ReLU(inplace=True),
                # Squeeze-and-Excite
                SELayer(hidden_dim) if use_se else nn.Identity(),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )
        else:
            self.conv = nn.Sequential(
                # pw
                nn.Conv2d(inp, hidden_dim, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hidden_dim),
                HSwish() if use_hs else nn.ReLU(inplace=True),
                # dw
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    kernel_size,
                    stride,
                    (kernel_size - 1) // 2,
                    groups=hidden_dim,
                    bias=False,
                ),
                nn.BatchNorm2d(hidden_dim),
                # Squeeze-and-Excite
                SELayer(hidden_dim) if use_se else nn.Identity(),
                HSwish() if use_hs else nn.ReLU(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )

    def forward(self, x):
        """Forward."""

        if self.identity:
            return x + self.conv(x)
        return self.conv(x)


class MobileNetV3Base(ModelInterface):
    """Base model of MobileNetV3."""

    def __init__(
        self,
        num_classes=1000,
        width_mult=1.0,
        in_channels=3,
        input_size=(224, 224),
        dropout_cls=None,
        pooling_type="avg",
        feature_dim=1280,
        IN_first=False,
        self_challenging_cfg=False,
        lr_finder=None,
        **kwargs,
    ):

        super().__init__(**kwargs)
        self.in_size = input_size
        self.num_classes = num_classes
        self.input_IN = nn.InstanceNorm2d(in_channels, affine=True) if IN_first else None
        self.pooling_type = pooling_type
        self.self_challenging_cfg = self_challenging_cfg
        self.width_mult = width_mult
        self.dropout_cls = dropout_cls
        self.lr_finder = lr_finder
        self.feature_dim = feature_dim

    def infer_head(self, x, skip_pool=False):
        """Inference head."""
        raise NotImplementedError

    def extract_features(self, x):
        """Extract features."""
        raise NotImplementedError

    def forward(self, x, return_featuremaps=False, get_embeddings=False, gt_labels=None):
        """Forward."""
        if self.input_IN is not None:
            x = self.input_IN(x)  # pylint: disable=not-callable

        y = self.extract_features(x)
        if return_featuremaps:
            return y
        # should be checked
        return y


class MobileNetV3(MobileNetV3Base):
    """MobileNetV3."""

    def __init__(self, cfgs, mode, IN_conv1=False, **kwargs):
        super().__init__(**kwargs)
        # setting of inverted residual blocks
        self.cfgs = cfgs
        assert mode in ["large", "small"]
        # building first layer
        input_channel = make_divisible(16 * self.width_mult, 8)
        stride = 1 if self.in_size[0] < 100 else 2
        layers = [conv_3x3_bn(3, input_channel, stride, IN_conv1)]
        # building inverted residual blocks
        block = InvertedResidual
        flag = True
        for k, t, c, use_se, use_hs, s in self.cfgs:
            if (self.in_size[0] < 100) and (s == 2) and flag:
                s = 1
                flag = False
            output_channel = make_divisible(c * self.width_mult, 8)
            exp_size = make_divisible(input_channel * t, 8)
            layers.append(block(input_channel, exp_size, output_channel, k, s, use_se, use_hs))
            input_channel = output_channel
        self.features = nn.Sequential(*layers)
        # building last several layers
        self.conv = conv_1x1_bn(input_channel, exp_size, self.loss)
        output_channel = {"large": 1280, "small": 1024}
        output_channel = (
            make_divisible(output_channel[mode] * self.width_mult, 8) if self.width_mult > 1.0 else output_channel[mode]
        )
        self._initialize_weights()

    def extract_features(self, x):
        """Extract features."""

        y = self.conv(self.features(x))
        return y

    def infer_head(self, x, skip_pool=False):
        """Inference head."""

        if not skip_pool:
            glob_features = self._glob_feature_vector(x, self.pooling_type, reduce_dims=False)
        else:
            glob_features = x

        logits = self.classifier(glob_features.view(x.shape[0], -1))
        return glob_features, logits

    def _initialize_weights(self):
        """Initialize weights."""

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()


@BACKBONES.register_module()
class OTXMobileNetV3(MobileNetV3):
    """MobileNetV3 model for OTX."""

    cfgs = dict(
        small=[
            # k, t, c, SE, HS, s
            [3, 1, 16, 1, 0, 2],
            [3, 4.5, 24, 0, 0, 2],
            [3, 3.67, 24, 0, 0, 1],
            [5, 4, 40, 1, 1, 2],
            [5, 6, 40, 1, 1, 1],
            [5, 6, 40, 1, 1, 1],
            [5, 3, 48, 1, 1, 1],
            [5, 3, 48, 1, 1, 1],
            [5, 6, 96, 1, 1, 2],
            [5, 6, 96, 1, 1, 1],
            [5, 6, 96, 1, 1, 1],
        ],
        large=[
            # k, t, c, SE, HS, s
            [3, 1, 16, 0, 0, 1],
            [3, 4, 24, 0, 0, 2],
            [3, 3, 24, 0, 0, 1],
            [5, 3, 40, 1, 0, 2],
            [5, 3, 40, 1, 0, 1],
            [5, 3, 40, 1, 0, 1],
            [3, 6, 80, 0, 1, 2],
            [3, 2.5, 80, 0, 1, 1],
            [3, 2.3, 80, 0, 1, 1],
            [3, 2.3, 80, 0, 1, 1],
            [3, 6, 112, 1, 1, 1],
            [3, 6, 112, 1, 1, 1],
            [5, 6, 160, 1, 1, 2],
            [5, 6, 160, 1, 1, 1],
            [5, 6, 160, 1, 1, 1],
        ],
    )

    def __init__(self, mode="large", width_mult=1.0, **kwargs):
        super().__init__(self.cfgs[mode], mode=mode, width_mult=width_mult, **kwargs)
        self.key = "mobilenetv3_" + mode
        if width_mult != 1.0:
            self.key = self.key + "_{:03d}".format(int(width_mult * 100))  # pylint: disable=consider-using-f-string
        self.init_weights(self.pretrained)

    def forward(self, x):
        """Forward."""

        return super().forward(x, return_featuremaps=True)

    def init_weights(self, pretrained=None):
        """Initialize weights."""

        if isinstance(pretrained, str) and os.path.exists(pretrained):
            load_checkpoint(self, pretrained)
            logger.info(f"init weight - {pretrained}")
        elif pretrained is not None:
            load_checkpoint(self, pretrained_urls[self.key])
            logger.info(f"init weight - {pretrained_urls[self.key]}")



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from functools import partial

from otx.algorithms.common.adapters.nncf.patches import (
    nncf_trace_wrapper,
    no_nncf_trace_wrapper,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmcv.nncf.test_helpers import create_model


@e2e_pytest_unit
def test_nncf_trace_context():
    model = create_model()
    bak = model.forward

    with model.nncf_trace_context({}, True):
        assert isinstance(model.forward, partial)
        assert model.forward.func == bak
    assert not isinstance(model.forward, partial)
    assert model.forward == bak

    with model.nncf_trace_context({}, False):
        assert isinstance(model.forward, partial)
        assert model.forward.func == model.forward_dummy
    assert not isinstance(model.forward, partial)
    assert model.forward == bak


@e2e_pytest_unit
def test_nncf_trace_wrapper():
    nncf_trace_wrapper("", lambda: None)


@e2e_pytest_unit
def test_no_nncf_trace_wrapper():
    no_nncf_trace_wrapper("", lambda: None)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field
from typing import List

import numpy as np
from torch.nn import functional as F

from .builder import OPS
from .movements import PadV1
from .op import Attribute, Operation


@dataclass
class InterpolateV4Attribute(Attribute):
    mode: str
    shape_calculation_mode: str
    coordinate_transformation_mode: str = field(default="half_pixel")
    nearest_mode: str = field(default="round_prefer_floor")
    antialias: bool = field(default=False)
    pads_begin: List[int] = field(default_factory=lambda: [0])
    pads_end: List[int] = field(default_factory=lambda: [0])
    cube_coeff: float = field(default=-0.75)

    def __post_init__(self):
        super().__post_init__()
        valid_mode = ["nearest", "linear", "linear_onnx", "cubic"]
        if self.mode not in valid_mode:
            raise ValueError(f"Invalid mode {self.mode}. " f"It must be one of {valid_mode}.")
        valid_shape_calculation_mode = ["sizes", "scales"]
        if self.shape_calculation_mode not in valid_shape_calculation_mode:
            raise ValueError(
                f"Invalid shape_calculation_mode {self.shape_calculation_mode}. "
                f"It must be one of {valid_shape_calculation_mode}."
            )
        valid_coordinate_transformation_mode = [
            "half_pixel",
            "pytorch_half_pixel",
            "asymmetric",
            "tf_half_pixel_for_nn",
            "align_corners",
        ]
        if self.coordinate_transformation_mode not in valid_coordinate_transformation_mode:
            raise ValueError(
                f"Invalid coordinate_transformation_mode {self.coordinate_transformation_mode}. "
                f"It must be one of {valid_coordinate_transformation_mode}."
            )
        valid_nearest_mode = [
            "round_prefer_floor",
            "round_prefer_ceil",
            "floor",
            "ceil",
            "simple",
        ]
        if self.nearest_mode not in valid_nearest_mode:
            raise ValueError(f"Invalid nearest_mode {self.nearest_mode}. " f"It must be one of {valid_nearest_mode}.")


@OPS.register()
class InterpolateV4(Operation[InterpolateV4Attribute]):
    TYPE = "Interpolate"
    VERSION = 4
    ATTRIBUTE_FACTORY = InterpolateV4Attribute

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pad = PadV1("tmp", shape=self.shape, pad_mode="constant")

    def forward(self, input, sizes, scales, axes=None):
        # TODO list
        # - handle 'linear_onnx' mode
        # - coordinate_transformation_mode
        # - nearest_mode
        # - cube_coeff
        # - antialias

        if axes is None:
            axes = list(range(input.dim()))
        else:
            axes = axes.detach().cpu().tolist()

        output = self.pad(input, self.attrs.pads_begin, self.attrs.pads_end, 0)

        mode = self.attrs.mode
        if mode == "linear" or mode == "linear_onnx":
            align_corners = False
            if output.dim() == 3:
                pass
            elif output.dim() == 4:
                mode = "bilinear"
            elif output.dim() == 5:
                mode = "trilinear"
        elif mode == "cubic":
            align_corners = False
            if output.dim() == 3:
                raise NotImplementedError
            elif output.dim() == 4:
                mode = "bicubic"
            elif output.dim() == 5:
                raise NotImplementedError
        elif mode == "nearest":
            align_corners = None
            pass
        else:
            raise NotImplementedError

        if self.attrs.shape_calculation_mode == "sizes":
            sizes = sizes.detach().cpu().numpy()
            sizes = sizes[np.argsort(axes)].tolist()
            if output.dim() == len(sizes):
                sizes = sizes[2:]

            return F.interpolate(
                input=output,
                size=sizes,
                scale_factor=None,
                mode=mode,
                align_corners=align_corners,
            )
        else:
            scales = scales.detach().cpu().numpy()
            scales = scales[np.argsort(axes)].tolist()
            if output.dim() == len(scales):
                scales = scales[2:]

            return F.interpolate(
                input=output,
                size=None,
                scale_factor=scales,
                mode=mode,
                align_corners=align_corners,
            )



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmseg.models import SEGMENTORS
from mmseg.models.segmentors.encoder_decoder import EncoderDecoder

from otx.mpa.deploy.utils import is_mmdeploy_enabled


@SEGMENTORS.register_module()
class OTXEncoderDecoder(EncoderDecoder):
    def simple_test(self, img, img_meta, rescale=True, output_logits=False):
        """Simple test with single image."""
        seg_logit = self.inference(img, img_meta, rescale)
        if output_logits:
            seg_pred = seg_logit
        else:
            if self.out_channels == 1:
                seg_pred = (seg_logit > self.decode_head.threshold).to(seg_logit).squeeze(1)
            else:
                seg_pred = seg_logit.argmax(dim=1)
        if torch.onnx.is_in_onnx_export():
            # our inference backend only support 4D output
            if seg_pred.dim() != 4:
                seg_pred = seg_pred.unsqueeze(0)
            return seg_pred
        seg_pred = seg_pred.cpu().numpy()
        seg_pred = list(seg_pred)
        return seg_pred


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER

    from otx.mpa.modules.hooks.recording_forward_hooks import FeatureVectorHook

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.segmentors.otx_encoder_decoder.OTXEncoderDecoder.extract_feat"
    )
    def single_stage_detector__extract_feat(ctx, self, img):
        feat = self.backbone(img)
        self.feature_map = feat
        if self.with_neck:
            feat = self.neck(feat)
        return feat

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.segmentors.otx_encoder_decoder.OTXEncoderDecoder.simple_test"
    )
    def single_stage_detector__simple_test(ctx, self, img, img_metas, **kwargs):
        # without output activation
        #  seg_logit = self.encode_decode(img, img_metas)
        #  feature_vector = FeatureVectorHook.func(self.feature_maps)
        #  return seg_logit, feature_vector

        # with output activation
        seg_logit = self.inference(img, img_metas, True)
        feature_vector = FeatureVectorHook.func(self.feature_map)
        return seg_logit, feature_vector



"""Resource manager class for HPO runner."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, Optional

import torch

from otx.algorithms.common.utils.utils import check_positive

logger = logging.getLogger(__name__)


class BaseResourceManager(ABC):
    """Abstract class for resource manager class."""

    @abstractmethod
    def reserve_resource(self, trial_id):
        """Reserve a resource."""
        raise NotImplementedError

    @abstractmethod
    def release_resource(self, trial_id):
        """Release a resource."""
        raise NotImplementedError

    @abstractmethod
    def have_available_resource(self):
        """Check that there is available resource."""
        raise NotImplementedError


class CPUResourceManager(BaseResourceManager):
    """Resource manager class for CPU.

    Args:
        num_parallel_trial (int, optional): How many trials to run in parallel. Defaults to 4.
    """

    def __init__(self, num_parallel_trial: int = 4):
        check_positive(num_parallel_trial, "num_parallel_trial")

        self._num_parallel_trial = num_parallel_trial
        self._usage_status: List = []

    def reserve_resource(self, trial_id: Any) -> Optional[Dict]:
        """Reserve a resource under 'trial_id'.

        Args:
            trial_id (Any): Name of trial to reserve the resource.

        Raises:
            RuntimeError: If there is already resource reserved by 'trial_id', then raise an error.
        """
        if not self.have_available_resource():
            return None
        if trial_id in self._usage_status:
            raise RuntimeError(f"{trial_id} already has reserved resource.")

        logger.debug(f"{trial_id} reserved.")
        self._usage_status.append(trial_id)
        return {}

    def release_resource(self, trial_id: Any):
        """Release a resource under 'trial_id'.

        Args:
            trial_id (Any): Name of trial which uses the resource to release.
        """
        if trial_id not in self._usage_status:
            logger.warning(f"{trial_id} trial don't use resource now.")
        else:
            self._usage_status.remove(trial_id)
            logger.debug(f"{trial_id} released.")

    def have_available_resource(self):
        """Check that there is available resource."""
        return len(self._usage_status) < self._num_parallel_trial


class GPUResourceManager(BaseResourceManager):
    """Resource manager class for GPU.

    Args:
        num_gpu_for_single_trial (int, optional): How many GPUs is used for a single trial. Defaults to 1.
        available_gpu (Optional[str], optional): How many GPUs are available. Defaults to None.
    """

    def __init__(self, num_gpu_for_single_trial: int = 1, available_gpu: Optional[str] = None):
        check_positive(num_gpu_for_single_trial, "num_gpu_for_single_trial")

        self._num_gpu_for_single_trial = num_gpu_for_single_trial
        self._available_gpu = self._set_available_gpu(available_gpu)
        self._usage_status: Dict[Any, List] = {}

    def _set_available_gpu(self, available_gpu: Optional[str] = None):
        if available_gpu is None:
            cuda_visible_devices = os.getenv("CUDA_VISIBLE_DEVICES")
            if cuda_visible_devices is not None:
                available_gpu_arr = self._transform_gpu_format_from_string_to_arr(cuda_visible_devices)
            else:
                num_gpus = torch.cuda.device_count()
                available_gpu_arr = list(range(num_gpus))
        else:
            available_gpu_arr = self._transform_gpu_format_from_string_to_arr(available_gpu)

        return available_gpu_arr

    def _transform_gpu_format_from_string_to_arr(self, gpu: str):
        for val in gpu.split(","):
            if not val.isnumeric():
                raise ValueError(
                    "gpu format is wrong. " "gpu should only have numbers delimited by ','.\n" f"your value is {gpu}"
                )
        return [int(val) for val in gpu.split(",")]

    def reserve_resource(self, trial_id: Any) -> Optional[Dict]:
        """Reserve a resource under 'trial_id'.

        Args:
            trial_id (Any): Name of trial to reserve the resource.

        Raises:
            RuntimeError: If there is already resource reserved by 'trial_id', then raise an error.

        Returns:
            Optional[Dict]: Training environment to use.
        """
        if not self.have_available_resource():
            return None
        if trial_id in self._usage_status:
            raise RuntimeError(f"{trial_id} already has reserved resource.")

        resource = list(self._available_gpu[: self._num_gpu_for_single_trial])
        self._available_gpu = self._available_gpu[self._num_gpu_for_single_trial :]

        self._usage_status[trial_id] = resource
        return {"CUDA_VISIBLE_DEVICES": ",".join([str(val) for val in resource])}

    def release_resource(self, trial_id: Any):
        """Release a resource under 'trial_id'.

        Args:
            trial_id (Any): Name of trial which uses the resource to release.
        """
        if trial_id not in self._usage_status:
            logger.warning(f"{trial_id} trial don't use resource now.")
        else:
            self._available_gpu.extend(self._usage_status[trial_id])
            del self._usage_status[trial_id]

    def have_available_resource(self):
        """Check that there is available resource."""
        return len(self._available_gpu) >= self._num_gpu_for_single_trial


def get_resource_manager(
    resource_type: Literal["gpu", "cpu"],
    num_parallel_trial: Optional[int] = None,
    num_gpu_for_single_trial: Optional[int] = None,
    available_gpu: Optional[str] = None,
) -> BaseResourceManager:
    """Get an appropriate resource manager depending on current environment.

    Args:
        resource_type (Literal["gpu", "cpu"]): Which type of resource to use.
                                               If can be changed depending on environment.
        num_parallel_trial (Optional[int]): How many trials to run in parallel. It's used for CPUResourceManager.
                                            Defaults to None.
        num_gpu_for_single_trial (Optional[int]): How many GPUs is used for a single trial.
                                                  It's used for GPUResourceManager. Defaults to None.
        available_gpu (Optional[str]): How many GPUs are available. It's used for GPUResourceManager. Defaults to None.

    Raises:
        ValueError: If resource_type is neither 'gpu' nor 'cpu', then raise an error.

    Returns:
        BaseResourceManager: Resource manager to use.
    """
    if resource_type == "gpu" and not torch.cuda.is_available():
        logger.warning("GPU can't be used now. resource type is modified to cpu.")
        resource_type = "cpu"

    if resource_type == "cpu":
        args = {"num_parallel_trial": num_parallel_trial}
        args = _remove_none_from_dict(args)
        return CPUResourceManager(**args)  # type: ignore
    if resource_type == "gpu":
        args = {"num_gpu_for_single_trial": num_gpu_for_single_trial, "available_gpu": available_gpu}  # type: ignore
        args = _remove_none_from_dict(args)
        return GPUResourceManager(**args)  # type: ignore
    raise ValueError(f"Available resource type is cpu, gpu. Your value is {resource_type}.")


def _remove_none_from_dict(dict_val: Dict):
    key_to_remove = [key for key, val in dict_val.items() if val is None]
    for key in key_to_remove:
        del dict_val[key]
    return dict_val



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import tempfile

import numpy as np
import torch
from nncf.torch.nncf_network import NNCFNetwork

from otx.algorithms.classification.adapters.mmcls.nncf.builder import (
    build_nncf_classifier,
)
from otx.algorithms.common.adapters.nncf.compression import NNCFMetaState
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmcv.nncf.test_helpers import (
    create_config,
    create_dataset,
    create_model,
)


@e2e_pytest_unit
def test_build_nncf_classifier():
    mock_config = create_config(lib="mmcls")
    model = create_model(lib="mmcls")
    create_dataset(lib="mmcls")

    with tempfile.TemporaryDirectory() as tempdir:
        model_path = os.path.join(tempdir, "model.bin")
        state_to_build = model.state_dict()
        torch.save(state_to_build, model_path)
        mock_config.load_from = model_path
        ctrl, model = build_nncf_classifier(mock_config)
        assert isinstance(model, NNCFNetwork)
        assert len([hook for hook in mock_config.custom_hooks if hook.type == "CompressionHook"]) == 1
        mock_config.pop("custom_hooks")

        torch.save(
            {
                "meta": {
                    "nncf_enable_compression": True,
                    "nncf_meta": NNCFMetaState(
                        data_to_build=np.zeros((50, 50, 3)),
                        compression_ctrl=ctrl.get_compression_state(),
                        state_to_build=state_to_build,
                    ),
                },
                "state_dict": model.state_dict(),
            },
            model_path,
        )
        ctrl, model = build_nncf_classifier(mock_config, model_path)
        assert isinstance(model, NNCFNetwork)
        assert len([hook for hook in mock_config.custom_hooks if hook.type == "CompressionHook"]) == 1



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from logging import Logger

import pytest
import torch.nn as nn
from mmcv.runner import EpochBasedRunner

from otx.algorithms.common.adapters.mmcv.hooks import (
    CancelTrainingHook,
    EarlyStoppingHook,
    EnsureCorrectBestCheckpointHook,
    OTXLoggerHook,
    OTXProgressHook,
    ReduceLROnPlateauLrUpdaterHook,
    StopLossNanTrainingHook,
)
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)

# TODO: Need to add EMAMomentumUpdateHook unit-test


class TestCancelTrainingHook:
    @e2e_pytest_unit
    def test_cancel_training_hook_initialization_params_validation(self):
        """
        <b>Description:</b>
        Check CancelTrainingHook object initialization parameters validation

        <b>Input data:</b>
        "interval" non-int type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        CancelTrainingHook object initialization parameter
        """
        with pytest.raises(ValueError):
            CancelTrainingHook(interval="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_cancel_training_hook_after_train_iter_params_validation(self):
        """
        <b>Description:</b>
        Check CancelTrainingHook object "after_train_iter" method input parameters validation

        <b>Input data:</b>
        CancelTrainingHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_train_iter" method
        """
        hook = CancelTrainingHook()
        with pytest.raises(ValueError):
            hook.after_train_iter(runner="unexpected string")  # type: ignore


class TestEnsureCorrectBestCheckpointHook:
    @e2e_pytest_unit
    def test_ensure_correct_best_checkpoint_hook_after_run_params_validation(self):
        """
        <b>Description:</b>
        Check EnsureCorrectBestCheckpointHook object "after_run" method input parameters validation

        <b>Input data:</b>
        EnsureCorrectBestCheckpointHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_run" method
        """
        hook = EnsureCorrectBestCheckpointHook()
        with pytest.raises(ValueError):
            hook.after_run(runner="unexpected string")  # type: ignore


class TestOTXLoggerHook:
    @e2e_pytest_unit
    def test_otx_logger_hook_initialization_parameters_validation(self):
        """
        <b>Description:</b>
        Check OTXLoggerHook object initialization parameters validation

        <b>Input data:</b>
        OTXLoggerHook object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        OTXLoggerHook object initialization parameter
        """
        correct_values_dict = {}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "curves" parameter
            ("curves", unexpected_str),
            # Unexpected string is specified as nested curve
            (
                "curves",
                {
                    "expected": OTXLoggerHook.Curve(),
                    "unexpected": unexpected_str,
                },
            ),
            # Unexpected string is specified as "interval" parameter
            ("interval", unexpected_str),
            # Unexpected string is specified as "ignore_last" parameter
            ("ignore_last", unexpected_str),
            # Unexpected string is specified as "reset_flag" parameter
            ("reset_flag", unexpected_str),
            # Unexpected string is specified as "by_epoch" parameter
            ("by_epoch", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=OTXLoggerHook,
        )

    @e2e_pytest_unit
    def test_otx_logger_hook_log_params_validation(self):
        """
        <b>Description:</b>
        Check OTXLoggerHook object "log" method input parameters validation

        <b>Input data:</b>
        OTXLoggerHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "log" method
        """
        hook = OTXLoggerHook()
        with pytest.raises(ValueError):
            hook.log(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_logger_hook_after_train_epoch_params_validation(self):
        """
        <b>Description:</b>
        Check OTXLoggerHook object "after_train_epoch" method input parameters validation

        <b>Input data:</b>
        OTXLoggerHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_train_epoch" method
        """
        hook = OTXLoggerHook()
        with pytest.raises(ValueError):
            hook.after_train_epoch(runner="unexpected string")  # type: ignore


class TestOTXProgressHook:
    @staticmethod
    def time_monitor():
        return TimeMonitorCallback(num_epoch=10, num_train_steps=5, num_test_steps=5, num_val_steps=4)

    def hook(self):
        return OTXProgressHook(time_monitor=self.time_monitor())

    @e2e_pytest_unit
    def test_otx_progress_hook_initialization_parameters_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object initialization parameters validation

        <b>Input data:</b>
        OTXProgressHook object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        OTXProgressHook object initialization parameter
        """
        correct_values_dict = {"time_monitor": self.time_monitor()}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "time_monitor" parameter
            ("time_monitor", unexpected_str),
            # Unexpected string is specified as "verbose" parameter
            ("verbose", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=OTXProgressHook,
        )

    @e2e_pytest_unit
    def test_otx_progress_hook_before_run_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "before_run" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "before_run" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.before_run(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_progress_hook_before_epoch_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "before_epoch" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "before_epoch" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.before_epoch(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_progress_hook_after_epoch_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "after_epoch" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_epoch" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.after_epoch(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_progress_hook_before_iter_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "before_iter" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "before_iter" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.before_iter(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_progress_hook_after_iter_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "after_iter" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_iter" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.after_iter(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_progress_hook_before_val_iter_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "before_val_iter" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "before_val_iter" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.before_val_iter(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_progress_hook_after_val_iter_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "after_val_iter" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_val_iter" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.after_val_iter(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_progress_hook_after_run_params_validation(self):
        """
        <b>Description:</b>
        Check OTXProgressHook object "after_run" method input parameters validation

        <b>Input data:</b>
        OTXProgressHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_run" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.after_run(runner="unexpected string")  # type: ignore


class TestEarlyStoppingHook:
    @staticmethod
    def hook():
        return EarlyStoppingHook(interval=5)

    @e2e_pytest_unit
    def test_early_stopping_hook_initialization_parameters_validation(self):
        """
        <b>Description:</b>
        Check EarlyStoppingHook object initialization parameters validation

        <b>Input data:</b>
        EarlyStoppingHook object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        EarlyStoppingHook object initialization parameter
        """
        correct_values_dict = {"interval": 5}
        unexpected_dict = {"unexpected": "dictionary"}
        unexpected_values = [
            # Unexpected dictionary is specified as "interval" parameter
            ("interval", unexpected_dict),
            # Unexpected dictionary is specified as "metric" parameter
            ("metric", unexpected_dict),
            # Unexpected dictionary is specified as "rule" parameter
            ("rule", unexpected_dict),
            # Unexpected dictionary is specified as "patience" parameter
            ("patience", unexpected_dict),
            # Unexpected dictionary is specified as "iteration_patience" parameter
            ("iteration_patience", unexpected_dict),
            # Unexpected dictionary is specified as "min_delta" parameter
            ("min_delta", unexpected_dict),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=EarlyStoppingHook,
        )

    @e2e_pytest_unit
    def test_early_stopping_hook_before_run_params_validation(self):
        """
        <b>Description:</b>
        Check EarlyStoppingHook object "before_run" method input parameters validation

        <b>Input data:</b>
        EarlyStoppingHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "before_run" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.before_run(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_early_stopping_hook_after_train_iter_params_validation(self):
        """
        <b>Description:</b>
        Check EarlyStoppingHook object "after_train_iter" method input parameters validation

        <b>Input data:</b>
        EarlyStoppingHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_train_iter" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.after_train_iter(runner="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_early_stopping_hook_after_train_epoch_params_validation(self):
        """
        <b>Description:</b>
        Check EarlyStoppingHook object "after_train_epoch" method input parameters validation

        <b>Input data:</b>
        EarlyStoppingHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_train_epoch" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.after_train_epoch(runner="unexpected string")  # type: ignore


class TestReduceLROnPlateauLrUpdaterHook:
    @staticmethod
    def hook():
        return ReduceLROnPlateauLrUpdaterHook(min_lr=0.1, interval=5)

    class MockModel(nn.Module):
        @staticmethod
        def train_step():
            pass

    @e2e_pytest_unit
    def test_reduce_lr_hook_initialization_parameters_validation(self):
        """
        <b>Description:</b>
        Check ReduceLROnPlateauLrUpdaterHook object initialization parameters validation

        <b>Input data:</b>
        ReduceLROnPlateauLrUpdaterHook object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        ReduceLROnPlateauLrUpdaterHook object initialization parameter
        """
        correct_values_dict = {"min_lr": 0.1, "interval": 5}
        unexpected_dict = {"unexpected": "dictionary"}
        unexpected_values = [
            # Unexpected dictionary is specified as "min_lr" parameter
            ("min_lr", unexpected_dict),
            # Unexpected dictionary is specified as "interval" parameter
            ("interval", unexpected_dict),
            # Unexpected dictionary is specified as "metric" parameter
            ("metric", unexpected_dict),
            # Unexpected dictionary is specified as "rule" parameter
            ("rule", unexpected_dict),
            # Unexpected dictionary is specified as "factor" parameter
            ("factor", unexpected_dict),
            # Unexpected dictionary is specified as "patience" parameter
            ("patience", unexpected_dict),
            # Unexpected dictionary is specified as "iteration_patience" parameter
            ("iteration_patience", unexpected_dict),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=ReduceLROnPlateauLrUpdaterHook,
        )

    @e2e_pytest_unit
    def test_reduce_lr_hook_get_lr_params_validation(self):
        """
        <b>Description:</b>
        Check ReduceLROnPlateauLrUpdaterHook object "get_lr" method input parameters validation

        <b>Input data:</b>
        ReduceLROnPlateauLrUpdaterHook object, "get_lr" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_lr" method
        """
        hook = self.hook()
        runner = EpochBasedRunner(model=self.MockModel(), logger=Logger(name="test logger"))
        correct_values_dict = {"runner": runner, "base_lr": 0.2}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "runner" parameter
            ("runner", unexpected_str),
            # Unexpected string is specified as "base_lr" parameter
            ("base_lr", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=hook.get_lr,
        )

    @e2e_pytest_unit
    def test_reduce_lr_hook_before_run_params_validation(self):
        """
        <b>Description:</b>
        Check ReduceLROnPlateauLrUpdaterHook object "before_run" method input parameters validation

        <b>Input data:</b>
        ReduceLROnPlateauLrUpdaterHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "before_run" method
        """
        hook = self.hook()
        with pytest.raises(ValueError):
            hook.before_run(runner="unexpected string")  # type: ignore


class TestStopLossNanTrainingHook:
    @e2e_pytest_unit
    def test_stop_loss_nan_train_hook_after_train_iter_params_validation(self):
        """
        <b>Description:</b>
        Check StopLossNanTrainingHook object "after_train_iter" method input parameters validation

        <b>Input data:</b>
        StopLossNanTrainingHook object, "runner" non-BaseRunner type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "after_train_iter" method
        """
        hook = StopLossNanTrainingHook()
        with pytest.raises(ValueError):
            hook.after_train_iter(runner="unexpected string")  # type: ignore



_base_ = [
    "./data.py",
    "./pipelines/ubt.py",
]

__dataset_type = "CocoDataset"
__data_root_path = "data/coco/"

__train_pipeline = {{_base_.train_pipeline}}
# __unlabeled_pipeline = __train_pipeline.copy().pop(1)  # Removing 'LoadAnnotations' op
__unlabeled_pipeline = {{_base_.unlabeled_pipeline}}
__test_pipeline = {{_base_.test_pipeline}}

__samples_per_gpu = 2

data = dict(
    samples_per_gpu=__samples_per_gpu,
    workers_per_gpu=2,
    train=dict(
        type=__dataset_type,
        ann_file=__data_root_path + "annotations/instances_train2017.json",
        img_prefix=__data_root_path + "train2017/",
        pipeline=__train_pipeline,
    ),
    unlabeled=dict(
        type=__dataset_type,
        img_file=None,
        img_prefix=__data_root_path + "train2017/",
        filter_empty_gt=False,
        pipeline=__unlabeled_pipeline,
    ),
    val=dict(
        type=__dataset_type,
        ann_file=__data_root_path + "annotations/instances_val2017.json",
        img_prefix=__data_root_path + "val2017/",
        test_mode=True,
        pipeline=__test_pipeline,
    ),
    test=dict(
        type=__dataset_type,
        ann_file=__data_root_path + "annotations/instances_val2017.json",
        img_prefix=__data_root_path + "val2017/",
        test_mode=True,
        pipeline=__test_pipeline,
    ),
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
from mmcv import ConfigDict

from otx.mpa.seg.stage import SegStage
from otx.mpa.utils.config_utils import update_or_add_custom_hook
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class IncrSegStage(SegStage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def configure_task(self, cfg, training, **kwargs):
        """Adjust settings for task adaptation"""
        super().configure_task(cfg, training, **kwargs)

        new_classes = np.setdiff1d(self.model_classes, self.org_model_classes).tolist()

        # FIXME : can be naive supervised learning (from-scratch ver.)
        # Check if new classes are added
        has_new_class = True if len(new_classes) > 0 else False
        if has_new_class is False:
            ValueError("Incremental learning should have at least one new class!")

        # Update TaskAdaptHook (use incremental sampler)
        task_adapt_hook = ConfigDict(
            type="TaskAdaptHook",
            src_classes=self.org_model_classes,
            dst_classes=self.model_classes,
            model_type=cfg.model.type,
            sampler_flag=has_new_class,
            efficient_mode=cfg["task_adapt"].get("efficient_mode", False),
        )
        update_or_add_custom_hook(cfg, task_adapt_hook)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.common.utils.mask_to_bbox import mask2bbox, mask_to_border
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestMask2Bbox:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.mask = np.zeros((3, 3))
        self.mask[0, 0] = 1

    @e2e_pytest_unit
    def test_mask_to_border(self):
        out = mask_to_border(self.mask)
        assert (out == self.mask).all()

    @e2e_pytest_unit
    def test_mask2bbox(self):
        out = mask2bbox(self.mask)
        expected_out = [[0, 0, 1, 1]]
        assert out == expected_out



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import time
from collections.abc import Mapping
from copy import deepcopy
from functools import partial
from subprocess import CalledProcessError
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import mmcv
import numpy as np
import onnx
import torch
from mmcv.parallel import collate, scatter

from .utils import numpy_2_list
from .utils.mmdeploy import (
    is_mmdeploy_enabled,
    mmdeploy_init_model_helper,
    update_deploy_cfg,
)
from .utils.onnx import prepare_onnx_for_openvino


class NaiveExporter:
    @staticmethod
    def export2openvino(
        output_dir: str,
        model_builder: Callable,
        cfg: mmcv.Config,
        input_data: Dict[Any, Any],
        *,
        precision: str = "FP32",
        model_name: str = "model",
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        opset_version: int = 11,
        dynamic_axes: Dict[Any, Any] = {},
        mo_transforms: str = "",
    ):
        input_data = scatter(collate([input_data], samples_per_gpu=1), [-1])[0]

        model = model_builder(cfg)
        model = model.cpu().eval()

        onnx_path = NaiveExporter.torch2onnx(
            output_dir,
            model,
            input_data,
            model_name=model_name,
            input_names=input_names,
            output_names=output_names,
            opset_version=opset_version,
            dynamic_axes=dynamic_axes,
        )

        def get_normalize_cfg(cfg):
            def _get_normalize_cfg(cfg_):
                out = []
                if isinstance(cfg_, Mapping):
                    for key, value in cfg_.items():
                        if isinstance(value, (Mapping, list)):
                            out += _get_normalize_cfg(value)
                        if key == "type" and value == "Normalize":
                            return [cfg_]
                elif isinstance(cfg_, list):
                    for value in cfg_:
                        if isinstance(value, (Mapping, list)):
                            out += _get_normalize_cfg(value)
                return out

            cfg = _get_normalize_cfg(cfg)
            assert len(cfg) == 1
            return cfg[0]

        mo_args = {}

        normalize_cfg = get_normalize_cfg(cfg.data.test)
        if normalize_cfg.get("mean", None) is not None:
            mo_args["mean_values"] = normalize_cfg.get("mean")
        if normalize_cfg.get("std", None) is not None:
            mo_args["scale_values"] = normalize_cfg.get("std")
        if normalize_cfg.get("to_rgb", False):
            mo_args["reverse_input_channels"] = None

        if precision == "FP16":
            mo_args["compress_to_fp16"] = None
        if mo_transforms:
            mo_args["transform"] = mo_transforms

        NaiveExporter.onnx2openvino(
            output_dir,
            onnx_path,
            model_name=model_name,
            **mo_args,
        )

    @staticmethod
    def torch2onnx(
        output_dir: str,
        model: torch.nn.Module,
        input_data: Dict[Any, Any],
        *,
        model_name: str = "model",
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        opset_version: int = 11,
        dynamic_axes: Dict[Any, Any] = {},
        verbose: bool = False,
        **onnx_options,
    ) -> str:

        img_metas = input_data.get("img_metas")
        numpy_2_list(img_metas)
        imgs = input_data.get("img")
        model.forward = partial(model.forward, img_metas=img_metas, return_loss=False)

        onnx_file_name = model_name + ".onnx"
        torch.onnx.export(
            model,
            imgs,
            os.path.join(output_dir, onnx_file_name),
            verbose=verbose,
            export_params=True,
            input_names=input_names,
            output_names=output_names,
            dynamic_axes=dynamic_axes,
            opset_version=opset_version,
            operator_export_type=torch.onnx.OperatorExportTypes.ONNX,
            **onnx_options,
        )
        model.__dict__.pop("forward")
        return os.path.join(output_dir, onnx_file_name)

    @staticmethod
    def onnx2openvino(
        output_dir: str,
        onnx_path: str,
        *,
        model_name: str = "model",
        **openvino_options,
    ) -> Tuple[str, str]:
        from otx.mpa.utils import mo_wrapper

        mo_args = {
            "input_model": onnx_path,
            "output_dir": output_dir,
            "model_name": model_name,
        }
        mo_args.update(openvino_options)

        ret, msg = mo_wrapper.generate_ir(output_dir, output_dir, silent=False, **mo_args)
        if ret != 0:
            raise ValueError(msg)
        return (
            os.path.join(output_dir, model_name + ".xml"),
            os.path.join(output_dir, model_name + ".bin"),
        )


if is_mmdeploy_enabled():
    import mmdeploy.apis.openvino as openvino_api
    from mmdeploy.apis import (
        build_task_processor,
        extract_model,
        get_predefined_partition_cfg,
        torch2onnx,
    )
    from mmdeploy.apis.openvino import get_input_info_from_cfg, get_mo_options_from_cfg
    from mmdeploy.core import FUNCTION_REWRITER
    from mmdeploy.utils import get_backend_config, get_ir_config, get_partition_config

    class MMdeployExporter:
        @staticmethod
        def export2openvino(
            output_dir: str,
            model_builder: Callable,
            cfg: mmcv.Config,
            deploy_cfg: mmcv.Config,
            *,
            model_name: str = "model",
        ):

            task_processor = build_task_processor(cfg, deploy_cfg, "cpu")

            def helper(*args, **kwargs):
                return mmdeploy_init_model_helper(*args, **kwargs, model_builder=model_builder)

            task_processor.__class__.init_pytorch_model = helper

            input_data_cfg = deploy_cfg.pop(
                "input_data",
                {"shape": (128, 128, 3), "file_path": None},
            )

            if input_data_cfg.get("file_path"):
                import cv2

                input_data = cv2.imread(input_data_cfg.get("file_path"))
                # image assumed to be RGB format under OTX
                input_data = cv2.cvtColor(input_data, cv2.COLOR_BGR2RGB)
            else:
                input_data = np.zeros(input_data_cfg["shape"], dtype=np.uint8)

            onnx_paths = []
            onnx_paths.append(
                MMdeployExporter.torch2onnx(
                    output_dir,
                    input_data,
                    cfg,
                    deploy_cfg,
                    model_name=model_name,
                )
            )

            partition_cfgs = get_partition_config(deploy_cfg)
            if partition_cfgs:
                partition_cfgs = partition_cfgs.get("partition_cfg", None)
                onnx_paths.extend(
                    MMdeployExporter.partition_onnx(
                        output_dir,
                        onnx_paths[0],
                        partition_cfgs,
                    )
                )

            for i, onnx_path in enumerate(onnx_paths):
                mo_options = {}
                if i > 0:
                    mo_options = partition_cfgs[i - 1].get("mo_options", {})
                deploy_cfg_ = deepcopy(deploy_cfg)
                update_deploy_cfg(onnx_path, deploy_cfg_, mo_options)
                MMdeployExporter.onnx2openvino(
                    output_dir,
                    onnx_path,
                    deploy_cfg_,
                )

        @staticmethod
        def torch2onnx(
            output_dir: str,
            input_data: Any,
            cfg: mmcv.Config,
            deploy_cfg: mmcv.Config,
            *,
            model_name: str = "model",
        ) -> str:
            onnx_file_name = model_name + ".onnx"
            torch2onnx(
                input_data,
                output_dir,
                onnx_file_name,
                deploy_cfg=deploy_cfg,
                model_cfg=cfg,
                model_checkpoint=cfg.get("load_from", None),
                device="cpu",
            )
            return os.path.join(output_dir, onnx_file_name)

        @staticmethod
        def partition_onnx(
            output_dir,
            onnx_path: str,
            partition_cfgs: Union[mmcv.ConfigDict, List[mmcv.ConfigDict]],
        ) -> Tuple[str, ...]:
            partitioned_paths = []

            if not isinstance(partition_cfgs, list):
                partition_cfgs = [partition_cfgs]

            for partition_cfg in partition_cfgs:
                save_file = partition_cfg["save_file"]
                save_path = os.path.join(output_dir, save_file)
                start = partition_cfg["start"]
                end = partition_cfg["end"]
                dynamic_axes = partition_cfg.get("dynamic_axes", None)

                extract_model(onnx_path, start, end, dynamic_axes=dynamic_axes, save_file=save_path)
                partitioned_paths.append(save_path)
            return tuple(partitioned_paths)

        @staticmethod
        def onnx2openvino(
            output_dir: str,
            onnx_path: str,
            deploy_cfg: Union[str, mmcv.Config],
            *,
            model_name: Optional[str] = None,
        ) -> Tuple[str, str]:

            input_info = get_input_info_from_cfg(deploy_cfg)
            output_names = get_ir_config(deploy_cfg).output_names
            mo_options = get_mo_options_from_cfg(deploy_cfg)

            if not model_name:
                model_name = os.path.basename(onnx_path).replace(".onnx", "")
            mo_options.args += f'--model_name "{model_name}" '

            onnx_ready_path = os.path.join(os.path.dirname(onnx_path), f"{model_name}_ready.onnx")
            prepare_onnx_for_openvino(onnx_path, os.path.join(os.path.dirname(onnx_path), f"{model_name}_ready.onnx"))

            try:
                openvino_api.from_onnx(onnx_ready_path, output_dir, input_info, output_names, mo_options)
            except CalledProcessError as e:
                # NOTE: mo returns non zero return code (245) even though it successfully generate IR
                cur_time = time.time()
                time_threshold = 5
                if not (
                    e.returncode == 245
                    and not {model_name + ".bin", model_name + ".xml"} - set(os.listdir(output_dir))
                    and (
                        os.path.getmtime(os.path.join(output_dir, model_name + ".bin")) - cur_time < time_threshold
                        and os.path.getmtime(os.path.join(output_dir, model_name + ".xml")) - cur_time < time_threshold
                    )
                ):
                    raise e

            return (
                os.path.join(output_dir, model_name + ".xml"),
                os.path.join(output_dir, model_name + ".bin"),
            )



"""Test cross focal loss."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest
import torch

from otx.mpa.modules.models.losses.cross_focal_loss import CrossSigmoidFocalLoss


class TestCrossFocalLoss:
    """Test cross focal loss."""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Create the loss object"""
        self.predictions = torch.tensor([[0, 1, 0], [0, 1, 0], [0, 0, 1]], dtype=torch.float32)
        self.labels = torch.tensor([1, 1, 2])
        self.loss = CrossSigmoidFocalLoss(num_classes=3)

    def test_loss_computation(self):
        """Test loss output."""
        assert np.round(self.loss(self.predictions, self.labels).item(), decimals=4) == 0.0885

    @pytest.mark.xfail(reason="This should fail as the masks are different")
    def test_loss_computation_with_mask(self):
        """Tests loss output with ignored label is provided."""
        valid_label_mask1 = torch.ones((3, 3))
        loss1 = self.loss(self.predictions, self.labels, valid_label_mask=valid_label_mask1)
        valid_label_mask2 = torch.zeros((3, 3))
        loss2 = self.loss(self.predictions, self.labels, valid_label_mask=valid_label_mask2)
        assert loss1 != loss2

    def test_reduction(self):
        """Test reduction."""

        loss1 = self.loss(self.predictions, self.labels, reduction_override="none")
        loss2 = self.loss(self.predictions, self.labels, reduction_override="mean")
        loss3 = self.loss(self.predictions, self.labels, reduction_override="sum")
        assert loss1.shape == (3, 3)
        assert loss2 != loss3



"""Model Wrapper for blur of OTX Segmentation."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import warnings
from typing import Any, Dict, Iterable, Optional, Union

import cv2
import numpy as np
from openvino.model_zoo.model_api.adapters.model_adapter import ModelAdapter
from openvino.model_zoo.model_api.models import SegmentationModel
from openvino.model_zoo.model_api.models.types import NumericalValue

from otx.api.utils.argument_checks import check_input_parameters_type
from otx.api.utils.segmentation_utils import create_hard_prediction_from_soft_prediction


@check_input_parameters_type()
def get_activation_map(features: Union[np.ndarray, Iterable, int, float]):
    """Getter activation_map functions."""
    min_soft_score = np.min(features)
    max_soft_score = np.max(features)
    factor = 255.0 / (max_soft_score - min_soft_score + 1e-12)

    float_act_map = factor * (features - min_soft_score)
    int_act_map = np.uint8(np.floor(float_act_map))
    int_act_map = cv2.applyColorMap(int_act_map, cv2.COLORMAP_JET)
    int_act_map = cv2.cvtColor(int_act_map, cv2.COLOR_BGR2RGB)
    return int_act_map


class BlurSegmentation(SegmentationModel):
    """BlurSegmentation class of openvino model wrapper."""

    __model__ = "blur_segmentation"

    @check_input_parameters_type()
    def __init__(self, model_adapter: ModelAdapter, configuration: Optional[dict] = None, preload: bool = False):
        super().__init__(model_adapter, configuration, preload)
        self.out_channels = 0

    @classmethod
    def parameters(cls):
        """BlurSegmentation.parameters function."""
        parameters = super().parameters()
        parameters.update(
            {
                "soft_threshold": NumericalValue(default_value=0.5, min=0.0, max=1.0),
                "blur_strength": NumericalValue(value_type=int, default_value=1, min=0, max=25),
            }
        )

        return parameters

    def _check_io_number(self, number_of_inputs, number_of_outputs):
        pass

    def _get_outputs(self):
        layer_name = "output"
        layer_shape = self.outputs[layer_name].shape

        if len(layer_shape) == 3:
            self.out_channels = 0
        elif len(layer_shape) == 4:
            self.out_channels = layer_shape[1]
        else:
            raise Exception(f"Unexpected output layer shape {layer_shape}. Only 4D and 3D output layers are supported")

        return layer_name

    @check_input_parameters_type()
    def postprocess(self, outputs: Dict[str, np.ndarray], meta: Dict[str, Any]):
        """BlurSegmentation.postprocess function."""
        predictions = outputs[self.output_blob_name].squeeze()
        soft_prediction = np.transpose(predictions, axes=(1, 2, 0))

        hard_prediction = create_hard_prediction_from_soft_prediction(
            soft_prediction=soft_prediction, soft_threshold=self.soft_threshold, blur_strength=self.blur_strength
        )
        hard_prediction = cv2.resize(
            hard_prediction, meta["original_shape"][1::-1], 0, 0, interpolation=cv2.INTER_NEAREST
        )
        soft_prediction = cv2.resize(
            soft_prediction, meta["original_shape"][1::-1], 0, 0, interpolation=cv2.INTER_NEAREST
        )
        meta["soft_prediction"] = soft_prediction

        if "feature_vector" not in outputs:
            warnings.warn("Could not find Feature Vector in OpenVINO output. Please rerun export or retrain the model.")
            meta["feature_vector"] = None
        else:
            meta["feature_vector"] = outputs["feature_vector"].reshape(-1)

        return hard_prediction



"""Tests for Action Detection Task with OTX CLI"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import pytest

from otx.api.entities.model_template import parse_model_template
from otx.cli.registry import Registry
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    otx_eval_openvino_testing,
    otx_eval_testing,
    otx_export_testing,
    otx_train_testing,
    pot_eval_testing,
    pot_optimize_testing,
)

# Finetuning arguments
# TODO: Need to change sample dataset
args = {
    "--train-data-roots": "tests/assets/cvat_dataset/action_detection/train",
    "--val-data-roots": "tests/assets/cvat_dataset/action_detection/train",
    "--test-data-roots": "tests/assets/cvat_dataset/action_detection/train",
    "train_params": ["params", "--learning_parameters.num_iters", "2", "--learning_parameters.batch_size", "4"],
}

otx_dir = os.getcwd()

TT_STABILITY_TESTS = os.environ.get("TT_STABILITY_TESTS", False)
if TT_STABILITY_TESTS:
    default_template = parse_model_template(
        os.path.join("otx/algorithms/action/configs", "detection", "x3d_fast_rcnn", "template.yaml")
    )
    templates = [default_template] * 100
    templates_ids = [template.model_template_id + f"-{i+1}" for i, template in enumerate(templates)]
else:
    templates = Registry("otx/algorithms/action").filter(task_type="ACTION_DETECTION").templates
    templates_ids = [template.model_template_id for template in templates]


class TestToolsOTXActionDetection:
    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "action_det"
        otx_train_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "action_det"
        otx_eval_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.skip(reason="CVS-102941 ONNX export of action detection model keeps failed")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_export(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "action_det"
        otx_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.skip(reason="CVS-102941 ONNX export of action detection model keeps failed")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "action_det"
        otx_eval_openvino_testing(template, tmp_dir_path, otx_dir, args, threshold=1.0)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.skip(reason="CVS-102941 ONNX export of action detection model keeps failed")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_optimize(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "action_det"
        pot_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.skip(reason="CVS-102941 ONNX export of action detection model keeps failed")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "action_det"
        pot_eval_testing(template, tmp_dir_path, otx_dir, args)



"""This module defines classes representing metadata information."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import abc
from enum import Enum, auto
from typing import Any, Optional

from otx.api.entities.model import ModelEntity


class IMetadata(metaclass=abc.ABCMeta):
    """This interface represents any additional metadata information which can be connected to an IMedia."""

    __name = Optional[str]

    @property
    def name(self):
        """Gets or sets the name of the Metadata item."""
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value


class FloatType(Enum):
    """Represents the use of the FloatMetadata."""

    FLOAT = auto()  # Regular float, without particular context
    EMBEDDING_VALUE = auto()
    ACTIVE_SCORE = auto()

    def __str__(self):
        """Return the name of FloatType enum."""
        return str(self.name)


class FloatMetadata(IMetadata):
    """This class represents metadata of type float.

    Args:
        name (str): Name of the metadata.
        value (float): Value of the metadata.
        float_type (FloatType): Type of the metadata.
    """

    def __init__(self, name: str, value: float, float_type: FloatType = FloatType.FLOAT):
        self.name = name
        self.value = value
        self.float_type = float_type

    def __repr__(self):
        """Prints the model, data and type of the MetadataItemEntity."""
        return f"FloatMetadata({self.name}, {self.value}, {self.float_type})"

    def __eq__(self, other):
        """Checks if two FloatMetadata have the same name, value and type."""
        return self.name == other.name and self.value == other.value and self.float_type == other.float_type


class VideoMetadata(IMetadata):
    """This class represents metadata of video.

    Args:
        video_id (str): id(name) for video.
        frame_idx (int): Index for frame.
        is_empty_frame(bool): whether this is empty frame(for action detection)
    """

    def __init__(self, video_id: str, frame_idx: int, is_empty_frame: bool):
        self.video_id = video_id
        self.frame_idx = frame_idx
        self.is_empty_frame = is_empty_frame
        self.metadata = {"video_id": video_id, "frame_idx": frame_idx, "is_empty_frame": is_empty_frame}

    def __repr__(self):
        """Prints the video_id, frame_id and type of the MetadataItemEntity."""
        out_string = "VideoMetadata"
        out_string += f"({self.metadata})"
        return out_string

    def __eq__(self, other):
        """Checks if two VideoMetadata have the same name, value and type."""
        return self.metadata == other.metadata

    def update(self, key: str, value: Any):
        """Update metadata infomation."""
        setattr(self, key, value)
        self.metadata[key] = value


class MetadataItemEntity:
    """This class is a wrapper class which connects the metadata value to model, which was used to generate it.

    Args:
        data (IMetadata): The metadata value.
        model (Optional[ModelEntity]): The model which was used to generate the metadata. Defaults to None.
    """

    def __init__(
        self,
        data: Any,
        model: Optional[ModelEntity] = None,
    ):
        self.data = data
        self.model = model

    def __repr__(self):
        """Prints the model and data of the MetadataItemEntity."""
        return f"MetadataItemEntity(model={self.model})"

    def __eq__(self, other):
        """Returns true if the model and the data match the other MetadataItemEntity."""
        return self.model == other.model and self.data == other.data



"""Collections of hooks for common OTX algorithms."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import math
import os
from collections import defaultdict
from math import cos, inf, isnan, pi
from typing import Any, Dict, List, Optional

from mmcv.parallel import is_module_wrapper
from mmcv.runner import BaseRunner, EpochBasedRunner
from mmcv.runner.dist_utils import master_only
from mmcv.runner.hooks import HOOKS, Hook, LoggerHook, LrUpdaterHook
from mmcv.utils import print_log

from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from otx.api.utils.argument_checks import check_input_parameters_type
from otx.mpa.utils.logger import get_logger

logger = get_logger()


# pylint: disable=too-many-instance-attributes, protected-access, too-many-arguments, unused-argument
@HOOKS.register_module()
class CancelTrainingHook(Hook):
    """CancelTrainingHook for Training Stopping."""

    @check_input_parameters_type()
    def __init__(self, interval: int = 5):
        """Periodically check whether whether a stop signal is sent to the runner during model training.

        Every 'check_interval' iterations, the work_dir for the runner is checked to see if a file '.stop_training'
        is present. If it is, training is stopped.

        :param interval: Period for checking for stop signal, given in iterations.

        """
        self.interval = interval

    @staticmethod
    def _check_for_stop_signal(runner: BaseRunner):
        """Log _check_for_stop_signal for CancelTrainingHook."""
        work_dir = runner.work_dir
        stop_filepath = os.path.join(work_dir, ".stop_training")
        if os.path.exists(stop_filepath):
            if isinstance(runner, EpochBasedRunner):
                epoch = runner.epoch
                runner._max_epochs = epoch  # Force runner to stop by pretending it has reached it's max_epoch
            runner.should_stop = True  # Set this flag to true to stop the current training epoch
            os.remove(stop_filepath)

    @check_input_parameters_type()
    def after_train_iter(self, runner: BaseRunner):
        """Log after_train_iter for CancelTrainingHook."""
        if not self.every_n_iters(runner, self.interval):
            return
        self._check_for_stop_signal(runner)


@HOOKS.register_module()
class EnsureCorrectBestCheckpointHook(Hook):
    """EnsureCorrectBestCheckpointHook.

    This hook makes sure that the 'best_mAP' checkpoint points properly to the best model, even if the best model is
    created in the last epoch.
    """

    @check_input_parameters_type()
    def after_run(self, runner: BaseRunner):
        """Called after train epoch hooks."""
        runner.call_hook("after_train_epoch")


@HOOKS.register_module()
class OTXLoggerHook(LoggerHook):
    """OTXLoggerHook for Logging."""

    class Curve:
        """Curve with x (epochs) & y (scores)."""

        def __init__(self):
            self.x = []
            self.y = []

        def __repr__(self):
            """Repr function."""
            points = []
            for x, y in zip(self.x, self.y):
                points.append(f"({x},{y})")
            return "curve[" + ",".join(points) + "]"

    @check_input_parameters_type()
    def __init__(
        self,
        curves: Optional[Dict[Any, Curve]] = None,
        interval: int = 10,
        ignore_last: bool = True,
        reset_flag: bool = True,
        by_epoch: bool = True,
    ):
        super().__init__(interval, ignore_last, reset_flag, by_epoch)
        self.curves = curves if curves is not None else defaultdict(self.Curve)

    @master_only
    @check_input_parameters_type()
    def log(self, runner: BaseRunner):
        """Log function for OTXLoggerHook."""
        tags = self.get_loggable_tags(runner, allow_text=False, tags_to_skip=())
        if runner.max_epochs is not None:
            normalized_iter = self.get_iter(runner) / runner.max_iters * runner.max_epochs
        else:
            normalized_iter = self.get_iter(runner)
        for tag, value in tags.items():
            curve = self.curves[tag]
            # Remove duplicates.
            if len(curve.x) > 0 and curve.x[-1] == normalized_iter:
                curve.x.pop()
                curve.y.pop()
            curve.x.append(normalized_iter)
            curve.y.append(value)

    @check_input_parameters_type()
    def after_train_epoch(self, runner: BaseRunner):
        """Called after_train_epoch in OTXLoggerHook."""
        # Iteration counter is increased right after the last iteration in the epoch,
        # temporarily decrease it back.
        runner._iter -= 1
        super().after_train_epoch(runner)
        runner._iter += 1


@HOOKS.register_module()
class OTXProgressHook(Hook):
    """OTXProgressHook for getting progress."""

    @check_input_parameters_type()
    def __init__(self, time_monitor: TimeMonitorCallback, verbose: bool = False):
        super().__init__()
        self.time_monitor = time_monitor
        self.verbose = verbose
        self.print_threshold = 1

    @check_input_parameters_type()
    def before_run(self, runner: BaseRunner):
        """Called before_run in OTXProgressHook."""
        total_epochs = runner.max_epochs if runner.max_epochs is not None else 1
        self.time_monitor.total_epochs = total_epochs
        self.time_monitor.train_steps = runner.max_iters // total_epochs if total_epochs else 1
        self.time_monitor.steps_per_epoch = self.time_monitor.train_steps + self.time_monitor.val_steps
        self.time_monitor.total_steps = max(math.ceil(self.time_monitor.steps_per_epoch * total_epochs), 1)
        self.time_monitor.current_step = 0
        self.time_monitor.current_epoch = 0
        self.time_monitor.on_train_begin()

    @check_input_parameters_type()
    def before_epoch(self, runner: BaseRunner):
        """Called before_epoch in OTXProgressHook."""
        self.time_monitor.on_epoch_begin(runner.epoch)

    @check_input_parameters_type()
    def after_epoch(self, runner: BaseRunner):
        """Called after_epoch in OTXProgressHook."""
        # put some runner's training status to use on the other hooks
        runner.log_buffer.output["current_iters"] = runner.iter
        self.time_monitor.on_epoch_end(runner.epoch, runner.log_buffer.output)

    @check_input_parameters_type()
    def before_iter(self, runner: BaseRunner):
        """Called before_iter in OTXProgressHook."""
        self.time_monitor.on_train_batch_begin(1)

    @check_input_parameters_type()
    def after_iter(self, runner: BaseRunner):
        """Called after_iter in OTXProgressHook."""
        # put some runner's training status to use on the other hooks
        runner.log_buffer.output["current_iters"] = runner.iter
        self.time_monitor.on_train_batch_end(1)
        if self.verbose:
            progress = self.progress
            if progress >= self.print_threshold:
                logger.warning(f"training progress {progress:.0f}%")
                self.print_threshold = (progress + 10) // 10 * 10

    @check_input_parameters_type()
    def before_val_iter(self, runner: BaseRunner):
        """Called before_val_iter in OTXProgressHook."""
        self.time_monitor.on_test_batch_begin(1, logger)

    @check_input_parameters_type()
    def after_val_iter(self, runner: BaseRunner):
        """Called after_val_iter in OTXProgressHook."""
        self.time_monitor.on_test_batch_end(1, logger)

    @check_input_parameters_type()
    def after_run(self, runner: BaseRunner):
        """Called after_run in OTXProgressHook."""
        self.time_monitor.on_train_end(1)
        if self.time_monitor.update_progress_callback:
            self.time_monitor.update_progress_callback(int(self.time_monitor.get_progress()))

    @property
    def progress(self):
        """Getting Progress from time monitor."""
        return self.time_monitor.get_progress()


@HOOKS.register_module()
class EarlyStoppingHook(Hook):
    """Cancel training when a metric has stopped improving.

    Early Stopping hook monitors a metric quantity and if no improvement is seen for a ‘patience’
    number of epochs, the training is cancelled.

    :param interval: the number of intervals for checking early stop. The interval number should be
                     the same as the evaluation interval - the `interval` variable set in
                     `evaluation` config.
    :param metric: the metric name to be monitored
    :param rule: greater or less.  In `less` mode, training will stop when the metric has stopped
                 decreasing and in `greater` mode it will stop when the metric has stopped
                 increasing.
    :param patience: Number of epochs with no improvement after which the training will be reduced.
                     For example, if patience = 2, then we will ignore the first 2 epochs with no
                     improvement, and will only cancel the training after the 3rd epoch if the
                     metric still hasn’t improved then
    :param iteration_patience: Number of iterations must be trained after the last improvement
                               before training stops. The same as patience but the training
                               continues if the number of iteration is lower than iteration_patience
                               This variable makes sure a model is trained enough for some
                               iterations after the last improvement before stopping.
    :param min_delta: Minimal decay applied to lr. If the difference between new and old lr is
                      smaller than eps, the update is ignored
    """

    rule_map = {"greater": lambda x, y: x > y, "less": lambda x, y: x < y}
    init_value_map = {"greater": -inf, "less": inf}
    greater_keys = [
        "acc",
        "top",
        "AR@",
        "auc",
        "precision",
        "mAP",
        "mDice",
        "mIoU",
        "mAcc",
        "aAcc",
    ]
    less_keys = ["loss"]

    @check_input_parameters_type()
    def __init__(
        self,
        interval: int,
        metric: str = "bbox_mAP",
        rule: Optional[str] = None,
        patience: int = 5,
        iteration_patience: int = 500,
        min_delta: float = 0.0,
    ):
        super().__init__()
        self.patience = patience
        self.iteration_patience = iteration_patience
        self.interval = interval
        self.min_delta = min_delta
        self._init_rule(rule, metric)

        self.min_delta *= 1 if self.rule == "greater" else -1
        self.last_iter = 0
        self.wait_count = 0
        self.by_epoch = True
        self.warmup_iters = 0
        self.best_score = self.init_value_map[self.rule]

    def _init_rule(self, rule, key_indicator):
        """Initialize rule, key_indicator, comparison_func, and best score.

        Here is the rule to determine which rule is used for key indicator
        when the rule is not specific:
        1. If the key indicator is in ``self.greater_keys``, the rule will be
           specified as 'greater'.
        2. Or if the key indicator is in ``self.less_keys``, the rule will be
           specified as 'less'.
        3. Or if the key indicator is equal to the substring in any one item
           in ``self.greater_keys``, the rule will be specified as 'greater'.
        4. Or if the key indicator is equal to the substring in any one item
           in ``self.less_keys``, the rule will be specified as 'less'.

        Args:
            rule (str | None): Comparison rule for best score.
            key_indicator (str | None): Key indicator to determine the
                comparison rule.
        """
        if rule not in self.rule_map and rule is not None:
            raise KeyError(f"rule must be greater, less or None, " f"but got {rule}.")

        if rule is None:
            if key_indicator in self.greater_keys or any(key in key_indicator for key in self.greater_keys):
                rule = "greater"
            elif key_indicator in self.less_keys or any(key in key_indicator for key in self.less_keys):
                rule = "less"
            else:
                raise ValueError(
                    f"Cannot infer the rule for key " f"{key_indicator}, thus a specific rule " f"must be specified."
                )
        self.rule = rule
        self.key_indicator = key_indicator
        self.compare_func = self.rule_map[self.rule]

    @check_input_parameters_type()
    def before_run(self, runner: BaseRunner):
        """Called before_run in EarlyStoppingHook."""
        self.by_epoch = runner.max_epochs is not None
        for hook in runner.hooks:
            if isinstance(hook, LrUpdaterHook):
                self.warmup_iters = hook.warmup_iters
                break

    @check_input_parameters_type()
    def after_train_iter(self, runner: BaseRunner):
        """Called after every training iter to evaluate the results."""
        if not self.by_epoch:
            self._do_check_stopping(runner)

    @check_input_parameters_type()
    def after_train_epoch(self, runner: BaseRunner):
        """Called after every training epoch to evaluate the results."""
        if self.by_epoch:
            self._do_check_stopping(runner)

    def _do_check_stopping(self, runner):
        """Called _do_check_stopping in EarlyStoppingHook."""
        if not self._should_check_stopping(runner) or self.warmup_iters > runner.iter:
            return

        if runner.rank == 0:
            if self.key_indicator not in runner.log_buffer.output:
                raise KeyError(
                    f"metric {self.key_indicator} does not exist in buffer. Please check "
                    f"{self.key_indicator} is cached in evaluation output buffer"
                )

            key_score = runner.log_buffer.output[self.key_indicator]
            if self.compare_func(key_score - self.min_delta, self.best_score):
                self.best_score = key_score
                self.wait_count = 0
                self.last_iter = runner.iter
            else:
                self.wait_count += 1
                if self.wait_count >= self.patience:
                    if runner.iter - self.last_iter < self.iteration_patience:
                        print_log(
                            f"\nSkip early stopping. Accumulated iteration "
                            f"{runner.iter - self.last_iter} from the last "
                            f"improvement must be larger than {self.iteration_patience} to trigger "
                            f"Early Stopping.",
                            logger=runner.logger,
                        )
                        return
                    stop_point = runner.epoch if self.by_epoch else runner.iter
                    print_log(
                        f"\nEarly Stopping at :{stop_point} with " f"best {self.key_indicator}: {self.best_score}",
                        logger=runner.logger,
                    )
                    runner.should_stop = True

    def _should_check_stopping(self, runner):
        """Called _should_check_stopping in EarlyStoppingHook."""
        check_time = self.every_n_epochs if self.by_epoch else self.every_n_iters
        if not check_time(runner, self.interval):
            # No evaluation during the interval.
            return False
        return True


@HOOKS.register_module(force=True)
class ReduceLROnPlateauLrUpdaterHook(LrUpdaterHook):
    """Reduce learning rate when a metric has stopped improving.

    Models often benefit from reducing the learning rate by a factor of 2-10 once learning stagnates.
    This scheduler reads a metrics quantity and if no improvement is seen for a ‘patience’
    number of epochs, the learning rate is reduced.

    :param min_lr: minimum learning rate. The lower bound of the desired learning rate.
    :param interval: the number of intervals for checking the hook. The interval number should be
                     the same as the evaluation interval - the `interval` variable set in
                     `evaluation` config.
    :param metric: the metric name to be monitored
    :param rule: greater or less.  In `less` mode, learning rate will be dropped if the metric has
                 stopped decreasing and in `greater` mode it will be dropped when the metric has
                 stopped increasing.
    :param patience: Number of epochs with no improvement after which learning rate will be reduced.
                     For example, if patience = 2, then we will ignore the first 2 epochs with no
                     improvement, and will only drop LR after the 3rd epoch if the metric still
                     hasn’t improved then
    :param iteration_patience: Number of iterations must be trained after the last improvement
                               before LR drops. The same as patience but the LR remains the same if
                               the number of iteration is lower than iteration_patience. This
                               variable makes sure a model is trained enough for some iterations
                               after the last improvement before dropping the LR.
    :param factor: Factor to be multiply with the learning rate.
                   For example, new_lr = current_lr * factor
    """

    rule_map = {"greater": lambda x, y: x > y, "less": lambda x, y: x < y}
    init_value_map = {"greater": -inf, "less": inf}
    greater_keys = [
        "acc",
        "top",
        "AR@",
        "auc",
        "precision",
        "mAP",
        "mDice",
        "mIoU",
        "mAcc",
        "aAcc",
    ]
    less_keys = ["loss"]

    @check_input_parameters_type()
    def __init__(
        self,
        min_lr: float,
        interval: int,
        metric: str = "bbox_mAP",
        rule: Optional[str] = None,
        factor: float = 0.1,
        patience: int = 3,
        iteration_patience: int = 300,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.interval = interval
        self.min_lr = min_lr
        self.factor = factor
        self.patience = patience
        self.iteration_patience = iteration_patience
        self.metric = metric
        self.bad_count = 0
        self.last_iter = 0
        self.current_lr = -1.0
        self.base_lr = []  # type: List
        self._init_rule(rule, metric)
        self.best_score = self.init_value_map[self.rule]

    def _init_rule(self, rule, key_indicator):
        """Initialize rule, key_indicator, comparison_func, and best score.

        Here is the rule to determine which rule is used for key indicator
        when the rule is not specific:
        1. If the key indicator is in ``self.greater_keys``, the rule will be
           specified as 'greater'.
        2. Or if the key indicator is in ``self.less_keys``, the rule will be
           specified as 'less'.
        3. Or if the key indicator is equal to the substring in any one item
           in ``self.greater_keys``, the rule will be specified as 'greater'.
        4. Or if the key indicator is equal to the substring in any one item
           in ``self.less_keys``, the rule will be specified as 'less'.

        Args:
            rule (str | None): Comparison rule for best score.
            key_indicator (str | None): Key indicator to determine the
                comparison rule.
        """
        if rule not in self.rule_map and rule is not None:
            raise KeyError(f"rule must be greater, less or None, " f"but got {rule}.")

        if rule is None:
            if key_indicator in self.greater_keys or any(key in key_indicator for key in self.greater_keys):
                rule = "greater"
            elif key_indicator in self.less_keys or any(key in key_indicator for key in self.less_keys):
                rule = "less"
            else:
                raise ValueError(
                    f"Cannot infer the rule for key " f"{key_indicator}, thus a specific rule " f"must be specified."
                )
        self.rule = rule
        self.key_indicator = key_indicator
        self.compare_func = self.rule_map[self.rule]

    def _is_check_timing(self, runner: BaseRunner) -> bool:
        """Check whether current epoch or iter is multiple of self.interval, skip during warmup interations."""
        check_time = self.after_each_n_epochs if self.by_epoch else self.after_each_n_iters
        return check_time(runner, self.interval) and (self.warmup_iters <= runner.iter)

    def after_each_n_epochs(self, runner: BaseRunner, interval: int) -> bool:
        """Check whether current epoch is a next epoch after multiples of interval."""
        return runner.epoch % interval == 0 if interval > 0 and runner.epoch != 0 else False

    def after_each_n_iters(self, runner: BaseRunner, interval: int) -> bool:
        """Check whether current iter is a next iter after multiples of interval."""
        return runner.iter % interval == 0 if interval > 0 and runner.iter != 0 else False

    @check_input_parameters_type()
    def get_lr(self, runner: BaseRunner, base_lr: float):
        """Called get_lr in ReduceLROnPlateauLrUpdaterHook."""
        if self.current_lr < 0:
            self.current_lr = base_lr

        if not self._is_check_timing(runner):
            return self.current_lr

        if hasattr(runner, "all_metrics"):
            score = runner.all_metrics.get(self.metric, 0.0)
        else:
            return self.current_lr

        if self.compare_func(score, self.best_score):
            self.best_score = score
            self.bad_count = 0
            self.last_iter = runner.iter
        else:
            self.bad_count += 1

        print_log(
            f"\nBest Score: {self.best_score}, Current Score: {score}, Patience: {self.patience} "
            f"Count: {self.bad_count}",
            logger=runner.logger,
        )

        if self.bad_count >= self.patience:
            if runner.iter - self.last_iter < self.iteration_patience:
                print_log(
                    f"\nSkip LR dropping. Accumulated iteration "
                    f"{runner.iter - self.last_iter} from the last "
                    f"improvement must be larger than {self.iteration_patience} to trigger "
                    f"LR dropping.",
                    logger=runner.logger,
                )
                return self.current_lr
            self.last_iter = runner.iter
            self.bad_count = 0
            print_log(
                f"\nDrop LR from: {self.current_lr}, to: " f"{max(self.current_lr * self.factor, self.min_lr)}",
                logger=runner.logger,
            )
            self.current_lr = max(self.current_lr * self.factor, self.min_lr)
        return self.current_lr

    @check_input_parameters_type()
    def before_run(self, runner: BaseRunner):
        """Called before_run in ReduceLROnPlateauLrUpdaterHook."""
        # TODO: remove overloaded method after fixing the issue
        #  https://github.com/open-mmlab/mmdetection/issues/6572
        for group in runner.optimizer.param_groups:
            group.setdefault("initial_lr", group["lr"])
        self.base_lr = [group["initial_lr"] for group in runner.optimizer.param_groups]
        self.bad_count = 0
        self.last_iter = 0
        self.current_lr = -1.0
        self.best_score = self.init_value_map[self.rule]


@HOOKS.register_module(force=True)
class StopLossNanTrainingHook(Hook):
    """StopLossNanTrainingHook."""

    @check_input_parameters_type()
    def after_train_iter(self, runner: BaseRunner):
        """Called after_train_iter in StopLossNanTrainingHook."""
        if isnan(runner.outputs["loss"].item()):
            logger.warning("Early Stopping since loss is NaN")
            runner.should_stop = True


@HOOKS.register_module()
class EMAMomentumUpdateHook(Hook):
    """Exponential moving average (EMA) momentum update hook for self-supervised methods.

    This hook includes momentum adjustment in self-supervised methods following:
        m = 1 - ( 1- m_0) * (cos(pi * k / K) + 1) / 2,
        k: current step, K: total steps.

    :param end_momentum: The final momentum coefficient for the target network, defaults to 1.
    :param update_interval: Interval to update new momentum, defaults to 1.
    :param by_epoch: Whether updating momentum by epoch or not, defaults to False.
    """

    def __init__(self, end_momentum: float = 1.0, update_interval: int = 1, by_epoch: bool = False, **kwargs):
        self.by_epoch = by_epoch
        self.end_momentum = end_momentum
        self.update_interval = update_interval

    def before_train_epoch(self, runner: BaseRunner):
        """Called before_train_epoch in EMAMomentumUpdateHook."""
        if not self.by_epoch:
            return

        if is_module_wrapper(runner.model):
            model = runner.model.module
        else:
            model = runner.model

        if not hasattr(model, "momentum"):
            raise AttributeError('The model must have attribute "momentum".')
        if not hasattr(model, "base_momentum"):
            raise AttributeError('The model must have attribute "base_momentum".')

        if self.every_n_epochs(runner, self.update_interval):
            cur_epoch = runner.epoch
            max_epoch = runner.max_epochs
            base_m = model.base_momentum
            updated_m = (
                self.end_momentum - (self.end_momentum - base_m) * (cos(pi * cur_epoch / float(max_epoch)) + 1) / 2
            )
            model.momentum = updated_m

    def before_train_iter(self, runner: BaseRunner):
        """Called before_train_iter in EMAMomentumUpdateHook."""
        if self.by_epoch:
            return

        if is_module_wrapper(runner.model):
            model = runner.model.module
        else:
            model = runner.model

        if not hasattr(model, "momentum"):
            raise AttributeError('The model must have attribute "momentum".')
        if not hasattr(model, "base_momentum"):
            raise AttributeError('The model must have attribute "base_momentum".')

        if self.every_n_iters(runner, self.update_interval):
            cur_iter = runner.iter
            max_iter = runner.max_iters
            base_m = model.base_momentum
            updated_m = (
                self.end_momentum - (self.end_momentum - base_m) * (cos(pi * cur_iter / float(max_iter)) + 1) / 2
            )
            model.momentum = updated_m

    def after_train_iter(self, runner: BaseRunner):
        """Called after_train_iter in EMAMomentumUpdateHook."""
        if self.every_n_iters(runner, self.update_interval):
            if is_module_wrapper(runner.model):
                runner.model.module.momentum_update()
            else:
                runner.model.momentum_update()


@HOOKS.register_module()
class ForceTrainModeHook(Hook):
    """Force train mode for model.

    This is a workaround of a bug in EvalHook from MMCV.
    If a model evaluation is enabled before training by setting 'start=0' in EvalHook,
    EvalHook does not put a model in a training mode again after evaluation.

    This simple hook forces to put a model in a training mode before every train epoch
    with the lowest priority.
    """

    def before_train_epoch(self, runner):
        """Make sure to put a model in a training mode before train epoch."""
        runner.model.train()


@HOOKS.register_module()
class TwoCropTransformHook(Hook):
    """TwoCropTransformHook with every specific interval.

    This hook decides whether using single pipeline or two pipelines
    implemented in `TwoCropTransform` for the current iteration.

    Args:
        interval (int): If `interval` == 1, both pipelines is used.
            If `interval` > 1, the first pipeline is used and then
            both pipelines are used every `interval`. Defaults to 1.
        by_epoch (bool): (TODO) Use `interval` by epoch. Defaults to False.
    """

    @check_input_parameters_type()
    def __init__(self, interval: int = 1, by_epoch: bool = False):
        assert interval > 0, f"interval (={interval}) must be positive value."
        if by_epoch:
            raise NotImplementedError("by_epoch is not implemented.")

        self.interval = interval
        self.cnt = 0

    @check_input_parameters_type()
    def _get_dataset(self, runner: BaseRunner):
        """Get dataset to handle `is_both`."""
        if hasattr(runner.data_loader.dataset, "dataset"):
            # for RepeatDataset
            dataset = runner.data_loader.dataset.dataset
        else:
            dataset = runner.data_loader.dataset

        return dataset

    # pylint: disable=inconsistent-return-statements
    @check_input_parameters_type()
    def _find_two_crop_transform(self, transforms: List[object]):
        """Find TwoCropTransform among transforms."""
        for transform in transforms:
            if transform.__class__.__name__ == "TwoCropTransform":
                return transform

    @check_input_parameters_type()
    def before_train_epoch(self, runner: BaseRunner):
        """Called before_train_epoch in TwoCropTransformHook."""
        # Always keep `TwoCropTransform` enabled.
        if self.interval == 1:
            return

        dataset = self._get_dataset(runner)
        two_crop_transform = self._find_two_crop_transform(dataset.pipeline.transforms)
        if self.cnt == self.interval - 1:
            # start using both pipelines
            two_crop_transform.is_both = True
        else:
            two_crop_transform.is_both = False

    @check_input_parameters_type()
    def after_train_iter(self, runner: BaseRunner):
        """Called after_train_iter in TwoCropTransformHook."""
        # Always keep `TwoCropTransform` enabled.
        if self.interval == 1:
            return

        if self.cnt < self.interval - 1:
            # Instead of using `runner.every_n_iters` or `runner.every_n_inner_iters`,
            # this condition is used to compare `self.cnt` with `self.interval` throughout the entire epochs.
            self.cnt += 1

        if self.cnt == self.interval - 1:
            dataset = self._get_dataset(runner)
            two_crop_transform = self._find_two_crop_transform(dataset.pipeline.transforms)
            if not two_crop_transform.is_both:
                # If `self.cnt` == `self.interval`-1, there are two cases,
                # 1. `self.cnt` was updated in L709, so `is_both` must be on for the next iter.
                # 2. if the current iter was already conducted, `is_both` must be off.
                two_crop_transform.is_both = True
            else:
                two_crop_transform.is_both = False
                self.cnt = 0



"""Inference Task of OTX Action Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import io
import os
import warnings
from typing import Dict, Iterable, Optional, Tuple

import numpy as np
import torch
from mmaction.datasets import build_dataloader, build_dataset
from mmaction.models import build_model
from mmaction.utils import get_root_logger
from mmcv.parallel import MMDataParallel
from mmcv.runner import load_checkpoint, load_state_dict
from mmcv.utils import Config

from otx.algorithms.action.adapters.mmaction import (
    export_model,
    patch_config,
    set_data_classes,
)
from otx.algorithms.action.configs.base import ActionConfig
from otx.algorithms.common.adapters.mmcv.utils import prepare_for_testing
from otx.algorithms.common.tasks.training_base import BaseTask
from otx.algorithms.common.utils.callback import InferenceProgressCallback
from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity, ModelFormat, ModelOptimizationType
from otx.api.entities.model_template import TaskType
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.entities.train_parameters import default_progress_callback
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType, IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.vis_utils import get_actmap

logger = get_root_logger()


# pylint: disable=too-many-locals, unused-argument
class ActionInferenceTask(BaseTask, IInferenceTask, IExportTask, IEvaluationTask, IUnload):
    """Inference Task Implementation of OTX Action Task."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment, **kwargs):
        super().__init__(ActionConfig, task_environment, **kwargs)

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main infer function of OTX Action Task."""
        logger.info("infer()")

        if inference_parameters:
            update_progress_callback = inference_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback

        self._time_monitor = InferenceProgressCallback(len(dataset), update_progress_callback)

        def pre_hook(module, inp):
            self._time_monitor.on_test_batch_begin(None, None)

        def hook(module, inp, out):
            self._time_monitor.on_test_batch_end(None, None)

        if self._recipe_cfg is None:
            self._init_task()
        if self._model:
            with self._model.register_forward_pre_hook(pre_hook), self._model.register_forward_hook(hook):
                prediction_results, _ = self._infer_model(dataset, inference_parameters)
            # TODO Load _add_predictions_to_dataset function from self._task_type
            if self._task_type == TaskType.ACTION_CLASSIFICATION:
                self._add_predictions_to_dataset(prediction_results, dataset)
            elif self._task_type == TaskType.ACTION_DETECTION:
                self._add_det_predictions_to_dataset(prediction_results, dataset)
            logger.info("Inference completed")
        else:
            raise Exception("Model initialization is failed")
        return dataset

    def _infer_model(
        self, dataset: DatasetEntity, inference_parameters: Optional[InferenceParameters] = None
    ) -> Tuple[Iterable, Optional[float]]:
        """Inference wrapper.

        This method triggers the inference and returns `prediction_results` zipped with prediction results,
        feature vectors, and saliency maps. `metric` is returned as a float value if InferenceParameters.is_evaluation
        is set to true, otherwise, `None` is returned.

        Args:
            dataset (DatasetEntity): the validation or test dataset to be inferred with
            inference_parameters (Optional[InferenceParameters], optional): Option to run evaluation or not.
                If `InferenceParameters.is_evaluation=True` then metric is returned, otherwise, both metric and
                saliency maps are empty. Defaults to None.

        Returns:
            Tuple[Iterable, float]: Iterable prediction results for each sample and metric for on the given dataset
        """
        if self._recipe_cfg is None:
            raise Exception("Recipe config is not initialized properly")

        dump_features = False
        dump_saliency_map = False

        test_config = prepare_for_testing(self._recipe_cfg, dataset)
        mm_test_dataset = build_dataset(test_config.data.test)
        # TODO Get batch size and num_gpus autometically
        batch_size = 1
        mm_test_dataloader = build_dataloader(
            mm_test_dataset,
            videos_per_gpu=batch_size,
            workers_per_gpu=test_config.data.workers_per_gpu,
            num_gpus=1,
            dist=False,
            shuffle=False,
        )

        self._model.eval()
        if torch.cuda.is_available():
            eval_model = MMDataParallel(self._model.cuda(test_config.gpu_ids[0]), device_ids=test_config.gpu_ids)
        else:
            eval_model = MMDataParallel(self._model)

        eval_predictions = []
        feature_vectors = []
        saliency_maps = []

        def dump_features_hook():
            raise NotImplementedError("get_feature_vector function for mmaction is not implemented")

        # pylint: disable=unused-argument
        def dummy_dump_features_hook(model, inp, out):
            feature_vectors.append(None)

        def dump_saliency_hook():
            raise NotImplementedError("get_saliency_map for mmaction is not implemented")

        # pylint: disable=unused-argument
        def dummy_dump_saliency_hook(model, inp, out):
            saliency_maps.append(None)

        feature_vector_hook = dump_features_hook if dump_features else dummy_dump_features_hook
        saliency_map_hook = dump_saliency_hook if dump_saliency_map else dummy_dump_saliency_hook

        # Use a single gpu for testing. Set in both mm_test_dataloader and eval_model
        with eval_model.module.backbone.register_forward_hook(feature_vector_hook):
            with eval_model.module.backbone.register_forward_hook(saliency_map_hook):
                for data in mm_test_dataloader:
                    with torch.no_grad():
                        result = eval_model(return_loss=False, **data)
                    eval_predictions.extend(result)

        # hard-code way to remove EvalHook args
        for key in ["interval", "tmpdir", "start", "gpu_collect", "save_best", "rule", "dynamic_intervals"]:
            self._recipe_cfg.evaluation.pop(key, None)

        metric = None
        metric_name = self._recipe_cfg.evaluation.final_metric
        if inference_parameters:
            if inference_parameters.is_evaluation:
                metric = mm_test_dataset.evaluate(eval_predictions, **self._recipe_cfg.evaluation)[metric_name]

        assert len(eval_predictions) == len(feature_vectors), f"{len(eval_predictions)} != {len(feature_vectors)}"
        assert len(eval_predictions) == len(saliency_maps), f"{len(eval_predictions)} != {len(saliency_maps)}"
        predictions = zip(eval_predictions, feature_vectors, saliency_maps)

        return predictions, metric

    # pylint: disable=attribute-defined-outside-init
    def _init_task(self, **kwargs):
        # FIXME: Temporary remedy for CVS-88098
        self._initialize(kwargs)
        logger.info(f"running task... kwargs = {kwargs}")
        if self._recipe_cfg is None:
            raise RuntimeError("'config' is not initialized yet. call prepare() method before calling this method")

        self._model = self._load_model(self._task_environment.model)

    def _load_model(self, model: ModelEntity):
        if self._recipe_cfg is None:
            raise Exception("Recipe config is not initialized properly")
        if model is not None:
            # If a model has been trained and saved for the task already, create empty model and load weights here
            buffer = io.BytesIO(model.get_data("weights.pth"))
            model_data = torch.load(buffer, map_location=torch.device("cpu"))

            self.confidence_threshold: float = model_data.get("confidence_threshold", self.confidence_threshold)
            model = self._create_model(self._recipe_cfg, from_scratch=True)

            try:
                load_state_dict(model, model_data["model"])

                # It prevent model from being overwritten
                if "load_from" in self._recipe_cfg:
                    self._recipe_cfg.load_from = None

                logger.info("Loaded model weights from Task Environment")
                logger.info(f"Model architecture: {self._model_name}")
                for name, weights in model.named_parameters():
                    if not torch.isfinite(weights).all():
                        logger.info(f"Invalid weights in: {name}. Recreate model from pre-trained weights")
                        model = self._create_model(self._recipe_cfg, from_scratch=False)
                        return model

            except BaseException as ex:
                raise ValueError("Could not load the saved model. The model file structure is invalid.") from ex
        else:
            # If there is no trained model yet, create model with pretrained weights as defined in the model config
            # file.
            model = self._create_model(self._recipe_cfg, from_scratch=False)
            logger.info(
                f"No trained model in project yet. Created new model with '{self._model_name}' "
                f"architecture and general-purpose pretrained weights."
            )
        return model

    @staticmethod
    def _create_model(config: Config, from_scratch: bool = False):
        """Creates a model, based on the configuration in config.

        :param config: mmaction configuration from which the model has to be built
        :param from_scratch: bool, if True does not load any weights

        :return model: ModelEntity in training mode
        """
        model_cfg = copy.deepcopy(config.model)

        init_from = None if from_scratch else config.get("load_from", None)
        logger.warning(init_from)
        if init_from is not None:
            # No need to initialize backbone separately, if all weights are provided.
            # model_cfg.pretrained = None
            logger.warning("build model")
            model = build_model(model_cfg)
            # Load all weights.
            logger.warning("load checkpoint")
            load_checkpoint(model, init_from, map_location="cpu")
        else:
            logger.warning("build model")
            model = build_model(model_cfg)
        return model

    @check_input_parameters_type()
    def evaluate(
        self,
        output_resultset: ResultSetEntity,
    ):
        """Evaluate function of OTX Action Task."""
        logger.info("called evaluate()")
        self._remove_empty_frames(output_resultset.ground_truth_dataset)
        metric = self._get_metric(output_resultset)
        performance = metric.get_performance()
        logger.info(f"Final model performance: {str(performance)}")
        output_resultset.performance = metric.get_performance()
        logger.info("Evaluation completed")

    def _get_metric(self, output_resultset: ResultSetEntity):
        if self._task_type == TaskType.ACTION_CLASSIFICATION:
            return MetricsHelper.compute_accuracy(output_resultset)
        if self._task_type == TaskType.ACTION_DETECTION:
            return MetricsHelper.compute_f_measure(output_resultset)
        raise NotImplementedError(f"{self._task_type} is not supported in action task")

    def _remove_empty_frames(self, dataset: DatasetEntity):
        """Remove empty frame for action detection dataset."""
        remove_indices = []
        for idx, item in enumerate(dataset):
            if item.get_metadata()[0].data.is_empty_frame:
                remove_indices.append(idx)
        dataset.remove_at_indices(remove_indices)

    def unload(self):
        """Unload the task."""
        if self._work_dir_is_temp:
            self._delete_scratch_space()

    @check_input_parameters_type()
    def export(self, export_type: ExportType, output_model: ModelEntity, dump_features: bool = True):
        """Export function of OTX Action Task."""
        # TODO: add dumping saliency maps and representation vectors according to dump_features flag
        if not dump_features:
            logger.warning(
                "Ommitting feature dumping is not implemented."
                "The saliency maps and representation vector outputs will be dumped in the exported model."
            )

        # copied from OTX inference_task.py
        logger.info("Exporting the model")
        if export_type != ExportType.OPENVINO:
            raise RuntimeError(f"not supported export type {export_type}")
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO
        self._init_task()

        try:
            from torch.jit._trace import TracerWarning

            warnings.filterwarnings("ignore", category=TracerWarning)
            export_model(
                self._model,
                self._recipe_cfg,
                onnx_model_path=f"{self._output_path}/openvino.onnx",
                output_dir_path=f"{self._output_path}",
            )
            bin_file = [f for f in os.listdir(self._output_path) if f.endswith(".bin")][0]
            xml_file = [f for f in os.listdir(self._output_path) if f.endswith(".xml")][0]
            with open(os.path.join(self._output_path, bin_file), "rb") as f:
                output_model.set_data("openvino.bin", f.read())
            with open(os.path.join(self._output_path, xml_file), "rb") as f:
                output_model.set_data("openvino.xml", f.read())
            output_model.set_data(
                "confidence_threshold", np.array([self.confidence_threshold], dtype=np.float32).tobytes()
            )
            output_model.precision = self._precision
            output_model.optimization_methods = self._optimization_methods
        except Exception as ex:
            raise RuntimeError("Optimization was unsuccessful.") from ex
        output_model.set_data("label_schema.json", label_schema_to_bytes(self._task_environment.label_schema))
        logger.info("Exporting completed")

    def _init_recipe_hparam(self) -> dict:
        configs = super()._init_recipe_hparam()
        configs.data.videos_per_gpu = configs.data.pop("samples_per_gpu", None)  # type: ignore[attr-defined]
        self._recipe_cfg.total_epochs = configs.runner.max_epochs  # type: ignore[attr-defined]
        # FIXME lr_config variables are hard-coded
        if hasattr(configs, "lr_config") and hasattr(configs["lr_config"], "warmup_iters"):
            self._recipe_cfg.lr_config.warmup = "linear"  # type: ignore[attr-defined]
            self._recipe_cfg.lr_config.warmup_by_epoch = True  # type: ignore[attr-defined]
        configs["use_adaptive_interval"] = self._hyperparams.learning_parameters.use_adaptive_interval
        return configs

    def _init_recipe(self):
        logger.info("called _init_recipe()")
        recipe_root = os.path.abspath(os.path.dirname(self.template_file_path))
        recipe = os.path.join(recipe_root, "model.py")
        self._recipe_cfg = Config.fromfile(recipe)
        patch_config(self._recipe_cfg, self.data_pipeline_path, self._output_path, self._task_type)
        set_data_classes(self._recipe_cfg, self._labels, self._task_type)
        logger.info(f"initialized recipe = {recipe}")

    def _init_model_cfg(self):
        model_cfg = Config.fromfile(os.path.join(self._model_dir, "model.py"))
        return model_cfg

    def _add_predictions_to_dataset(self, prediction_results: Iterable, dataset: DatasetEntity):
        """Loop over dataset again to assign predictions. Convert from MM format to OTX format."""
        prediction_results = list(prediction_results)
        video_info: Dict[str, int] = {}
        for dataset_item in dataset:
            video_id = dataset_item.get_metadata()[0].data.video_id
            if video_id not in video_info:
                video_info[video_id] = len(video_info)
        for dataset_item in dataset:
            video_id = dataset_item.get_metadata()[0].data.video_id
            all_results, feature_vector, saliency_map = prediction_results[video_info[video_id]]
            item_labels = []
            label = ScoredLabel(label=self._labels[all_results.argmax()], probability=all_results.max())
            item_labels.append(label)
            dataset_item.append_labels(item_labels)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                saliency_map = get_actmap(saliency_map, (dataset_item.width, dataset_item.height))
                saliency_map_media = ResultMediaEntity(
                    name="Saliency Map",
                    type="saliency_map",
                    annotation_scene=dataset_item.annotation_scene,
                    numpy=saliency_map,
                    roi=dataset_item.roi,
                )
                dataset_item.append_metadata_item(saliency_map_media, model=self._task_environment.model)

    def _add_det_predictions_to_dataset(self, prediction_results: Iterable, dataset: DatasetEntity):
        confidence_threshold = 0.05
        self._remove_empty_frames(dataset)
        for dataset_item, (all_results, feature_vector, saliency_map) in zip(dataset, prediction_results):
            shapes = []
            for label_idx, detections in enumerate(all_results):
                for i in range(detections.shape[0]):
                    probability = float(detections[i, 4])
                    coords = detections[i, :4]

                    if probability < confidence_threshold:
                        continue
                    if coords[3] - coords[1] <= 0 or coords[2] - coords[0] <= 0:
                        continue

                    assigned_label = [ScoredLabel(self._labels[label_idx], probability=probability)]
                    shapes.append(
                        Annotation(
                            Rectangle(x1=coords[0], y1=coords[1], x2=coords[2], y2=coords[3]),
                            labels=assigned_label,
                        )
                    )
            dataset_item.append_annotations(shapes)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                saliency_map = get_actmap(saliency_map, (dataset_item.width, dataset_item.height))
                saliency_map_media = ResultMediaEntity(
                    name="Saliency Map",
                    type="saliency_map",
                    annotation_scene=dataset_item.annotation_scene,
                    numpy=saliency_map,
                    roi=dataset_item.roi,
                )
                dataset_item.append_metadata_item(saliency_map_media, model=self._task_environment.model)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcls.models.builder import HEADS
from mmcls.models.heads.linear_head import LinearClsHead

from otx.mpa.modules.models.heads.non_linear_cls_head import NonLinearClsHead


class SemiClsHead:
    """Classification head for Semi-SL.

    Args:
        unlabeled_coef (float): unlabeled loss coefficient, default is 1.0
        dynamic_threshold (boolean): whether to use dynamic threshold, default is True
        min_threshold (float): Minimum value of threshold determining pseudo-label, default is 0.5
    """

    def __init__(self, unlabeled_coef=1.0, use_dynamic_threshold=True, min_threshold=0.5):
        self.unlabeled_coef = unlabeled_coef
        self.use_dynamic_threshold = use_dynamic_threshold
        self.min_threshold = (
            min_threshold if self.use_dynamic_threshold else 0.95
        )  # the range of threshold will be [min_thr, 1.0]
        self.num_pseudo_label = 0
        self.classwise_acc = torch.ones((self.num_classes,)) * self.min_threshold
        if torch.cuda.is_available():
            self.classwise_acc = self.classwise_acc.cuda()

    def loss(self, logits, gt_label, pseudo_label=None, mask=None):
        """loss function in which unlabeled data is considered

        Args:
            logit (set): (labeled data logit, unlabeled data logit)
            gt_label (Tensor): target features for labeled data
            pseudo_label (Tensor): target feature for unlabeled data
            mask (Tensor): Mask that shows pseudo-label that passes threshold

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        logits_x, logits_u_s = logits
        num_samples = len(logits_x)
        losses = dict()

        # compute supervised loss
        lx = self.compute_loss(logits_x, gt_label, avg_factor=num_samples)

        lu = 0
        if len(logits_u_s) > 0:
            # compute unsupervised loss
            lu = self.compute_loss(logits_u_s, pseudo_label, avg_factor=len(logits_u_s)) * mask
        losses["loss"] = lx + self.unlabeled_coef * lu
        losses["unlabeled_loss"] = self.unlabeled_coef * lu

        # compute accuracy
        acc = self.compute_accuracy(logits_x, gt_label)
        losses["accuracy"] = {f"top-{k}": a for k, a in zip(self.topk, acc)}
        return losses

    def forward_train(self, x, gt_label, final_layer=None):
        """forward_train head using pseudo-label selected through threshold

        Args:
            x (dict or Tensor): dict(labeled, unlabeled_weak, unlabeled_strong) or NxC input features.
            gt_label (Tensor): NxC target features.
            final_layer (nn.Linear or nn.Sequential): a final layer forwards feature from backbone.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        label_u, mask = None, None
        if isinstance(x, dict):
            outputs = final_layer(x["labeled"])  # Logit of Labeled Img
            batch_size = len(outputs)

            with torch.no_grad():
                logit_uw = final_layer(x["unlabeled_weak"])
                pseudo_label = torch.softmax(logit_uw.detach(), dim=-1)
                max_probs, label_u = torch.max(pseudo_label, dim=-1)

                # select Pseudo-Label using flexible threhold
                if torch.cuda.is_available():
                    max_probs = max_probs.cuda()
                mask = max_probs.ge(self.classwise_acc[label_u]).float()
                self.num_pseudo_label = mask.sum()

                if self.use_dynamic_threshold:
                    # get Labeled Data True Positive Confidence
                    logit_x = torch.softmax(outputs.detach(), dim=-1)
                    x_probs, x_idx = torch.max(logit_x, dim=-1)
                    x_probs = x_probs[x_idx == gt_label]
                    x_idx = x_idx[x_idx == gt_label]

                    # get Unlabeled Data Selected Confidence
                    uw_probs = max_probs[mask == 1]
                    uw_idx = label_u[mask == 1]

                    # update class-wise accuracy
                    for i in set(x_idx.tolist() + uw_idx.tolist()):
                        current_conf = torch.tensor([x_probs[x_idx == i].mean(), uw_probs[uw_idx == i].mean()])
                        current_conf = current_conf[~current_conf.isnan()].mean()
                        self.classwise_acc[i] = max(current_conf, self.min_threshold)

            outputs = torch.cat((outputs, final_layer(x["unlabeled_strong"])))
        else:
            outputs = final_layer(x)
            batch_size = len(outputs)

        logits_x = outputs[:batch_size]
        logits_u = outputs[batch_size:]
        del outputs
        logits = (logits_x, logits_u)
        losses = self.loss(logits, gt_label, label_u, mask)
        return losses


@HEADS.register_module()
class SemiLinearClsHead(SemiClsHead, LinearClsHead):
    """Linear classification head for Semi-SL

    This head is designed to support FixMatch algorithm. (https://arxiv.org/abs/2001.07685)
        - [OTX] supports dynamic threshold based on confidence for each class

    Args:
        num_classes (int): The number of classes of dataset used for training
        in_channels (int): The channels of input data from classifier
        loss (dict): configuration of loss, default is CrossEntropyLoss
        topk (set): evaluation topk score, default is (1, )
        unlabeled_coef (float): unlabeled loss coefficient, default is 1.0
        dynamic_threshold (boolean): whether to use dynamic threshold, default is True
        min_threshold (float): Minimum value of threshold determining pseudo-label, default is 0.5
    """

    def __init__(
        self,
        num_classes,
        in_channels,
        loss=dict(type="CrossEntropyLoss", loss_weight=1.0),
        topk=(1,),
        unlabeled_coef=1.0,
        use_dynamic_threshold=True,
        min_threshold=0.5,
    ):
        if in_channels <= 0:
            raise ValueError(f"in_channels={in_channels} must be a positive integer")
        if num_classes <= 0:
            raise ValueError("at least one class must be exist num_classes.")

        topk = (1,) if num_classes < 5 else (1, 5)
        LinearClsHead.__init__(self, num_classes, in_channels, loss=loss, topk=topk)
        SemiClsHead.__init__(self, unlabeled_coef, use_dynamic_threshold, min_threshold)

    def forward_train(self, x, gt_label):
        return SemiClsHead.forward_train(self, x, gt_label, final_layer=self.fc)


@HEADS.register_module()
class SemiNonLinearClsHead(SemiClsHead, NonLinearClsHead):
    """Non-linear classification head for Semi-SL

    This head is designed to support FixMatch algorithm. (https://arxiv.org/abs/2001.07685)
        - [OTX] supports dynamic threshold based on confidence for each class

    Args:
        num_classes (int): The number of classes of dataset used for training
        in_channels (int): The channels of input data from classifier
        hid_channels (int): Number of channels of hidden layer.
        act_cfg (dict): Config of activation layer.
        loss (dict): configuration of loss, default is CrossEntropyLoss
        topk (set): evaluation topk score, default is (1, )
        unlabeled_coef (float): unlabeled loss coefficient, default is 1.0
        dynamic_threshold (boolean): whether to use dynamic threshold, default is True
        min_threshold (float): Minimum value of threshold determining pseudo-label, default is 0.5
    """

    def __init__(
        self,
        num_classes,
        in_channels,
        hid_channels=1280,
        act_cfg=dict(type="ReLU"),
        loss=dict(type="CrossEntropyLoss", loss_weight=1.0),
        topk=(1,),
        dropout=False,
        unlabeled_coef=1.0,
        use_dynamic_threshold=True,
        min_threshold=0.5,
    ):
        if in_channels <= 0:
            raise ValueError(f"in_channels={in_channels} must be a positive integer")
        if num_classes <= 0:
            raise ValueError("at least one class must be exist num_classes.")

        topk = (1,) if num_classes < 5 else (1, 5)
        NonLinearClsHead.__init__(
            self,
            num_classes,
            in_channels,
            hid_channels=hid_channels,
            act_cfg=act_cfg,
            loss=loss,
            topk=topk,
            dropout=dropout,
        )
        SemiClsHead.__init__(self, unlabeled_coef, use_dynamic_threshold, min_threshold)

    def forward_train(self, x, gt_label):
        return SemiClsHead.forward_train(self, x, gt_label, final_layer=self.classifier)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import re
from dataclasses import dataclass, fields
from typing import Generic, Optional, Tuple, Type, TypeVar, Union

import torch

from ..utils import get_op_name
from .utils import get_dynamic_shape


@dataclass
class Attribute:
    shape: Optional[Union[Tuple[Tuple[int]], Tuple[int]]]

    def __post_init__(self):
        if self.shape is not None and not isinstance(self.shape, tuple):
            raise ValueError("shape must be a tuple of ints or a tuple of tuples of ints.")


_T = TypeVar("_T", bound=Attribute)


class Operation(torch.nn.Module, Generic[_T]):
    TYPE = ""
    VERSION = -1
    ATTRIBUTE_FACTORY: Type[_T] = Attribute

    def __init__(self, name: str, **kwargs):
        super().__init__()
        self._name = name
        self._attrs = self.ATTRIBUTE_FACTORY(**kwargs)

    @classmethod
    def from_ov(cls, ov_op):
        op_type = ov_op.get_type_name()
        op_version = ov_op.get_version()
        op_name = get_op_name(ov_op)
        assert cls.TYPE != "" and cls.VERSION >= 0
        assert op_type == cls.TYPE
        assert op_version == cls.VERSION

        attrs = ov_op.get_attributes()
        if "shape" not in attrs:
            shapes = []
            for output in ov_op.outputs():
                shapes.append(get_dynamic_shape(output))
            shapes = tuple(tuple(shape) for shape in shapes)
            attrs["shape"] = shapes

        return cls(name=op_name, **attrs)

    @property
    def type(self) -> str:
        return self.TYPE

    @property
    def version(self) -> int:
        return self.VERSION

    @property
    def name(self) -> str:
        return self._name

    @property
    def attrs(self):
        return self._attrs

    @property
    def shape(self) -> Optional[Union[Tuple[Tuple[int]], Tuple[int]]]:
        return self.attrs.shape
        #  shape = self.attrs.get("shape", None)
        #  if shape is not None and len(shape) == 1:
        #      shape = shape[0]
        #  return shape

    def __repr__(self):
        repr = f"{self.__class__.__name__}("
        repr += f"name={self.name}, "
        for field in fields(self.attrs):
            key = field.name
            if key == "shape":
                continue
            value = getattr(self.attrs, key)
            repr += f"{key}={value}, "
        repr = re.sub(", $", "", repr)
        repr += ")"
        return repr



"""Unit Test for otx.algorithms.action.utils.data."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np

from otx.algorithms.action.utils.data import (
    find_label_by_name,
    load_cls_annotations,
    load_cls_dataset,
    load_det_annotations,
    load_det_dataset,
)
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import generate_labels


@e2e_pytest_unit
def test_find_label_by_name() -> None:
    labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
    assert find_label_by_name(labels, "1", Domain.ACTION_CLASSIFICATION).name == "1"
    assert find_label_by_name(labels, "5", Domain.ACTION_CLASSIFICATION).name == "5"


@e2e_pytest_unit
def test_load_cls_annotations(mocker) -> None:
    class MockFile:
        def __init__(self, *args, **kwargs):
            self.lines = [
                "# Comment\n",
                "vid_0 5 1\n",
                "vid_1 5 1\n",
                "vid_2 5 1\n",
                "vid_3 5 2\n",
                "vid_4 5 2\n",
            ]

        def __enter__(self):
            return self.lines

        def __exit__(self, *args, **kwargs):
            return None

    mocker.patch("otx.algorithms.action.utils.data.open", return_value=MockFile())

    video_infos = load_cls_annotations("ann_file", "data_root")
    assert len(video_infos) == 5
    assert video_infos[0]["frame_dir"] == "data_root/vid_0"
    assert video_infos[0]["total_frames"] == 5
    assert video_infos[0]["label"] == 1


@e2e_pytest_unit
def test_load_det_annotations(mocker) -> None:
    class MockFile:
        def __init__(self, *args, **kwargs):
            self.lines = [
                "vid_0,0,0,0,1,1,1,0\n",
                "vid_0,1,0,0,1,1,1,1\n",
                "vid_0,1,0,0,1,1,2,1\n",
                "vid_0,2,0,0,1,1,1,2\n",
                "vid_0,3,0,0,1,1,2,3\n",
                "vid_0,4,0,0,1,1,2,4\n",
            ]

        def __enter__(self):
            return self.lines

        def __exit__(self, *args, **kwargs):
            return None

    mocker.patch("otx.algorithms.action.utils.data.open", return_value=MockFile())
    video_infos = load_det_annotations("ann_file", "data_root")
    assert len(video_infos) == 5
    assert video_infos[1]["frame_dir"] == "data_root/vid_0"
    assert video_infos[1]["video_id"] == "vid_0"
    assert video_infos[1]["timestamp"] == 1
    assert video_infos[1]["img_key"] == "vid_0,1"
    assert np.all(video_infos[1]["ann"]["gt_bboxes"] == np.array([[0.0, 0.0, 1.0, 1.0]]))
    assert np.all(video_infos[1]["ann"]["gt_labels"] == np.array([1, 2]))
    assert np.all(video_infos[1]["ann"]["entity_ids"] == np.array([1]))


@e2e_pytest_unit
def test_load_cls_dataset(mocker) -> None:
    """Test load_cls_dataset function."""

    mocker.patch(
        "otx.algorithms.action.utils.data.load_cls_annotations",
        return_value=[{"frame_dir": "data_root/vid_0", "total_frames": 5, "label": 1}],
    )
    mocker.patch(
        "otx.algorithms.action.utils.data.find_label_by_name",
        return_value=LabelEntity(name="1", domain=Domain.ACTION_CLASSIFICATION, id=ID(1)),
    )
    items = load_cls_dataset("ann_file", "data_root", Domain.ACTION_CLASSIFICATION)
    assert len(items) == 1
    assert items[0].media == {"frame_dir": "data_root/vid_0", "total_frames": 5}
    assert items[0].annotation_scene.get_labels()[0].name == "1"


@e2e_pytest_unit
def test_load_det_dataset(mocker) -> None:
    """Test load_det_dataset function."""

    mocker.patch(
        "otx.algorithms.action.utils.data.load_det_annotations",
        return_value=[
            {
                "frame_dir": "data_root/vid_0",
                "video_id": "vid_0",
                "timestamp": 0,
                "img_key": "vid_0,0",
                "ann": {
                    "gt_bboxes": np.array([[0.0, 0.0, 1.0, 1.0]]),
                    "gt_labels": [np.array([1])],
                    "entity_ids": np.array([0]),
                },
                "width": 320,
                "height": 240,
            }
        ],
    )
    mocker.patch(
        "otx.algorithms.action.utils.data.find_label_by_name",
        return_value=LabelEntity(name="1", domain=Domain.ACTION_DETECTION, id=ID(1)),
    )
    items = load_det_dataset("ann_file", "data_root", Domain.ACTION_DETECTION)
    assert len(items) == 1
    assert items[0].media == {
        "frame_dir": "data_root/vid_0",
        "video_id": "vid_0",
        "timestamp": 0,
        "img_key": "vid_0,0",
        "width": 320,
        "height": 240,
    }
    assert items[0].annotation_scene.get_labels()[0].name == "1"



"""OTX Adapters - mmaction.utils."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .config_utils import patch_config, prepare_for_training, set_data_classes
from .det_eval_utils import det_eval
from .export_utils import export_model

__all__ = ["patch_config", "set_data_classes", "prepare_for_training", "det_eval", "export_model"]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch

#  from mmdet.models.necks.ssd_neck import SSDNeck
import torch.nn as nn
from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule
from mmcv.runner import BaseModule
from mmdet.models.builder import NECKS

from ...mmov_model import MMOVModel


# FIXME: get rid of defined SSDNeck as this is a workaround for forked/deprecated mmdet
class SSDNeck(BaseModule):
    """Extra layers of SSD backbone to generate multi-scale feature maps.

    Args:
        in_channels (Sequence[int]): Number of input channels per scale.
        out_channels (Sequence[int]): Number of output channels per scale.
        level_strides (Sequence[int]): Stride of 3x3 conv per level.
        level_paddings (Sequence[int]): Padding size of 3x3 conv per level.
        l2_norm_scale (float|None): L2 normalization layer init scale.
            If None, not use L2 normalization on the first input feature.
        last_kernel_size (int): Kernel size of the last conv layer.
            Default: 3.
        use_depthwise (bool): Whether to use DepthwiseSeparableConv.
            Default: False.
        conv_cfg (dict): Config dict for convolution layer. Default: None.
        norm_cfg (dict): Dictionary to construct and config norm layer.
            Default: None.
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='ReLU').
        init_cfg (dict or list[dict], optional): Initialization config dict.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        level_strides,
        level_paddings,
        l2_norm_scale=20.0,
        last_kernel_size=3,
        use_depthwise=False,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type="ReLU"),
        init_cfg=[
            dict(type="Xavier", distribution="uniform", layer="Conv2d"),
            dict(type="Constant", val=1, layer="BatchNorm2d"),
        ],
    ):
        super(SSDNeck, self).__init__(init_cfg)
        assert len(out_channels) > len(in_channels)
        assert len(out_channels) - len(in_channels) == len(level_strides)
        assert len(level_strides) == len(level_paddings)
        assert in_channels == out_channels[: len(in_channels)]

        if l2_norm_scale:
            self.l2_norm = L2Norm(in_channels[0], l2_norm_scale)
            self.init_cfg += [dict(type="Constant", val=self.l2_norm.scale, override=dict(name="l2_norm"))]

        self.extra_layers = nn.ModuleList()
        extra_layer_channels = out_channels[len(in_channels) :]
        second_conv = DepthwiseSeparableConvModule if use_depthwise else ConvModule

        for i, (out_channel, stride, padding) in enumerate(zip(extra_layer_channels, level_strides, level_paddings)):
            kernel_size = last_kernel_size if i == len(extra_layer_channels) - 1 else 3
            per_lvl_convs = nn.Sequential(
                ConvModule(
                    out_channels[len(in_channels) - 1 + i],
                    out_channel // 2,
                    1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                ),
                second_conv(
                    out_channel // 2,
                    out_channel,
                    kernel_size,
                    stride=stride,
                    padding=padding,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                ),
            )
            self.extra_layers.append(per_lvl_convs)

    def forward(self, inputs):
        """Forward function."""
        outs = [feat for feat in inputs]
        if hasattr(self, "l2_norm"):
            outs[0] = self.l2_norm(outs[0])

        feat = outs[-1]
        for layer in self.extra_layers:
            feat = layer(feat)
            outs.append(feat)
        return tuple(outs)


class L2Norm(nn.Module):
    def __init__(self, n_dims, scale=20.0, eps=1e-10):
        """L2 normalization layer.

        Args:
            n_dims (int): Number of dimensions to be normalized
            scale (float, optional): Defaults to 20..
            eps (float, optional): Used to avoid division by zero.
                Defaults to 1e-10.
        """
        super(L2Norm, self).__init__()
        self.n_dims = n_dims
        self.weight = nn.Parameter(torch.Tensor(self.n_dims))
        self.eps = eps
        self.scale = scale

    def forward(self, x):
        """Forward function."""
        # normalization layer convert to FP32 in FP16 training
        x_float = x.float()
        norm = x_float.pow(2).sum(1, keepdim=True).sqrt() + self.eps
        return (self.weight[None, :, None, None].float().expand_as(x_float) * x_float / norm).type_as(x)


@NECKS.register_module()
class MMOVSSDNeck(SSDNeck):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        init_weight: bool = False,
        verify_shape: bool = True,
    ):
        # dummy
        in_channels = (512, 1024)
        out_channels = (512, 1024, 512, 256, 256, 256)
        level_strides = (2, 2, 1, 1)
        level_paddings = (1, 1, 0, 0)
        l2_norm_scale = None
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            level_strides=level_strides,
            level_paddings=level_paddings,
            l2_norm_scale=l2_norm_scale,
        )

        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._init_weight = init_weight

        self.extra_layers = torch.nn.ModuleList()

        for input, output in zip(inputs["extra_layers"], outputs["extra_layers"]):
            if input and output:
                self.extra_layers.append(
                    MMOVModel(
                        model_path_or_model,
                        weight_path,
                        inputs=input,
                        outputs=output,
                        remove_normalize=False,
                        merge_bn=False,
                        paired_bn=False,
                        init_weight=init_weight,
                        verify_shape=verify_shape,
                    )
                )
            else:
                self.extra_layers.append(torch.nn.Identity())

        if "l2_norm" in inputs and "l2_norm" in outputs:
            for input, output in zip(inputs["l2_norm"], outputs["l2_norm"]):
                self.l2_norm = MMOVModel(
                    model_path_or_model,
                    weight_path,
                    inputs=input,
                    outputs=output,
                    remove_normalize=False,
                    merge_bn=False,
                    paired_bn=False,
                    init_weight=init_weight,
                    verify_shape=verify_shape,
                )

    def init_weights(self, pretrained=None):
        # TODO
        pass
